# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import yaml
from collections import defaultdict
from metagenomix import parameters
from metagenomix.parameters import *


class Soft(object):

    def __init__(self, config):
        self.name = ''
        self.prev = None
        self.scratch = None   # no use of the scratch file system by default
        self.params = dict(config.params)  # init with default params
        # key attributes to be filled by each tool-specific code
        self.io = {}
        self.inputs = {}
        self.outputs = {}
        self.cmds = {}
        self.dirs = set()
        self.defaults = {}

    def get_softs(self, softs):
        if len(softs) == 1:
            self.name = softs[0]
        else:
            self.prev, self.name = softs


class Graph(object):
    """https://www.geeksforgeeks.org/find-paths-given-source-destination/"""
    def __init__(self, vertices):
        self.V = vertices
        self.graph = defaultdict(list)
        self.paths = defaultdict(list)

    def add_edge(self, u, v):
        self.graph[u].append(v)

    def print_paths_util(self, u, d, visited, path):
        visited[u] = True
        path.append(u)
        if u == d:
            self.paths[path[-1]].append(list(path))
        else:
            for i in self.graph[u]:
                if not visited[i]:
                    self.print_paths_util(i, d, visited, path)
        path.pop()
        visited[u] = False

    def print_paths(self, s, d):
        visited = [False] * self.V
        path = []
        self.print_paths_util(s, d, visited, path)


class Workflow(object):
    """Collect the data associated with each dataset passed but the user
    """
    def __init__(self, config, databases) -> None:
        self.config = config
        self.databases = databases
        self.graph = None
        self.names = ['fastq']
        self.softs = {}
        self.names_idx = {}
        self.names_idx_rev = {}
        self.skip = {}

    def run(self) -> None:
        """
        Parse the list of softwares, collect their sequence in the
        attribute `self.softs`, and collect all the paths of their
        sequences in the object `self.graph.paths`.
        """
        for softs in self.config.pipeline:
            self.collect_soft_name(softs)
            soft = Soft(self.config)
            soft.get_softs(softs)
            self.check_workflow(softs)
            self.softs[softs[-1]] = soft
        self.get_names_idx()
        self.make_graph()
        self.get_paths()
        self.set_params()

    def check_workflow(self, softs):
        if softs[-1] in self.softs:
            already_used = self.softs[softs[-1]]
            prev, name = already_used.prev, already_used.name
            message = 'Error in workflow "%s":\n' % self.config.pipeline_tsv
            message += "\tCan't run \"%s\" after \"%s\": " % tuple(softs[::-1])
            message += '"%s" already planned to use after "%s"\n' % (name, prev)
            message += "Each tool must yield a single output re-used as input\n"
            message += '-> please re-run for each different workflow "graph"\n'
            message += "Exiting\n"
            sys.exit(message)

    def collect_soft_name(self, softs: list) -> None:
        """Collect the sequential list of softwares.

        Parameters
        ----------
        softs : list
            One or two names of softwares that are run on
            the fastq files or after one another, respectively.
        """
        if softs[-1] not in self.names:
            if len(softs) > 1 and softs[0] not in self.names:
                raise IOError('"%s" not planned before "%s"' % tuple(softs))
            self.names.append(softs[-1])

    def get_names_idx(self) -> None:
        """
        Get a sequential numeric index (key) per software (value) in invoked
        order and the reversed, i.e., software (key) per numeric index (value).
        """
        self.names_idx = {y: x for x, y in enumerate(self.names)}
        self.names_idx_rev = {x: y for x, y in enumerate(self.names)}

    def make_graph(self) -> None:
        """
        Make the graph of all possible sequences of the
        list of softwares.
        """
        self.graph = Graph(len(self.names))
        for idx, soft in self.softs.items():
            if soft.prev:
                self.graph.add_edge(self.names_idx[soft.prev],
                                    self.names_idx[soft.name])
            else:
                self.graph.add_edge(0, self.names_idx[soft.name])

    def get_paths(self) -> None:
        """
        Collect all the paths of their sequences in the
        object `self.graph.paths`.
        """
        for soft in self.names[1:]:
            self.graph.print_paths(0, self.names_idx[soft])
        self.graph.paths = {self.names_idx_rev[idx]: [
            [self.names_idx_rev[p] for p in path] for path in paths]
            for idx, paths in self.graph.paths.items()}

    def make_dirs(self):
        for name, soft in self.softs.items():
            for directory in sorted(soft.dirs):
                if not isdir(directory):
                    os.makedirs(directory)

    def check_basic_params(self, user_params, soft):
        ints = ['time', 'procs', 'mem_num', 'chunks']
        for param, value in user_params.items():
            if param in ints:
                check_int(param, value, soft.name)
            elif param == 'mem_dim':
                check_mems(param, value, soft.name)
            elif param == 'env':
                check_env(self.config, value, soft.name)
            elif param == 'path':
                check_path(self.config, value, soft.name)
            elif param == 'scratch':
                check_scratch(value, soft.name)
            soft.params[param] = value

    def ignored_params(self, user_params, soft):
        valid_params = set(soft.defaults) | set(self.config.params)
        for param in sorted(user_params):
            if param not in valid_params:
                name = soft.name
                print('[%s] Param "%s" unknown and ignored' % (name, param))
                del user_params[param]

    def set_scratch(self, soft):
        """scratch set on command line overrides per-software scratches"""
        if self.config.localscratch:
            soft.params['scratch'] = self.config.localscratch
        elif self.config.scratch:
            soft.params['scratch'] = 'scratch'
        elif self.config.userscratch:
            soft.params['scratch'] = 'userscratch'

    def set_user_params(self, soft):
        name = soft.name.split('_')[0]
        user_params = dict(self.config.user_params.get(name, {}))
        func = 'check_%s' % name
        if hasattr(parameters, func) and callable(getattr(parameters, func)):
            check_ = getattr(parameters, func)
            soft.defaults = check_(self, user_params, soft)
            self.ignored_params(user_params, soft)

        self.check_basic_params(user_params, soft)
        if isinstance(soft.params['scratch'], int):
            soft.params['mem_num'] = soft.params['scratch']
            soft.params['mem_dim'] = 'gb'

    def skip_params(self, soft) -> bool:
        if '_' in soft.name:
            if self.skip.get(soft.name.split('_')[0], False):
                return True
            self.skip[soft.name.split('_')[0]] = True

    def show_params(self, soft):
        params_show = dict(
            x for x in soft.params.items() if x[0] not in self.config.params)
        if params_show:
            print('[Parameters] %s' % soft.name)
            print('=' * 30)
            if soft.name.startswith('search'):
                databases = params_show['databases']
                del params_show['databases']
                params = {'search': {soft.name.split('_')[-1]: params_show,
                                     'databases': databases}}
                print(yaml.dump(params))
            else:
                print(yaml.dump(params_show))
            print('%s defaults %s' % ('-' * 10, '-' * 10))
            print(yaml.dump(soft.defaults))
            print('=' * 30)

    def set_params(self) -> None:
        """
        Update the default params assigned to each software with the
        params passed by the user for each of the software.
        """
        for _, soft in self.softs.items():
            # print()
            # print()
            # print('-----------------')
            # print(soft.name)
            self.set_scratch(soft)
            self.set_user_params(soft)
            # print('soft.params:', soft.params)
            # print('-----------------')
        if self.config.show_params:
            for _, soft in self.softs.items():
                self.show_params(soft)

    # def collect_paths(self):
    #
    #     if soft.soft_prev == 'qiita_wol':
    #         soft.output_paths[(None, 'qiita_wol')] = qiita_wol
    #         if soft == 'instrain':
    #             instrain_refs, instrain_bams = get_instrain_refs_bams(soft,
    #                                                                   soft_prev)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import glob
import subprocess
import pkg_resources
from os.path import exists, isdir, isfile, splitext
from metagenomix._io_utils import read_yaml, fill_fastq
from metagenomix._metadata import read_metadata
import pandas as pd
import numpy as np

RESOURCES = pkg_resources.resource_filename("metagenomix", "resources")


class AnalysesConfig(object):
    """Collect the data associated with each dataset passed but the user."""
    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)
        self.meta = pd.DataFrame()
        self.pooling_groups = []
        self.conda_envs = {}
        self.modules = {}
        self.soft_paths = []
        self.fastq_scratch = {}
        self.fastq = {}
        self.params = {}
        self.dir = ''
        self.r = {}
        # self.cazy_focus_dbs = {}  <---- see databases "self.cazys"
        self.instrain = {'refs': {}, 'bams': {}}

    def run(self):
        self.check_xhpc_install()
        self.get_conda_envs()
        self.set_metadata()
        self.set_fastq()
        self.get_r()
        self.set_output()
        self.parse_yamls()
        self.set_coassembly()
        self.update_metadata()
        self.get_default_params()

    def check_xhpc_install(self):
        """Try to get the install path of third party tool
        [Xhpc](https://github.com/FranckLejzerowicz/Xhpc).
            If it exists, nothing happens and the code proceeds.
            Otherwise, the code ends and tells what to do.
        """
        if self.jobs:
            ret, _ = subprocess.getstatusoutput('which Xhpc')
            if ret:
                raise IOError('Xhpc not installed')
            else:
                xpbs_config = subprocess.getoutput('Xhpc --show-config')
                if not xpbs_config.startswith('* email address config file'):
                    raise IOError('Need to run Xhpc to setup config file')

    def get_conda_envs(self):
        """Get the names of the conda environments."""
        for env in subprocess.getoutput('conda env list').split('\n'):
            if env.startswith('#') or '*' in env or not env.strip():
                continue
            name, path = env.split()
            self.conda_envs[name] = path

    def set_metadata(self):
        """Read metadata with first column as index."""
        if not isfile(self.meta_fp):
            raise IOError('No file "%s"' % self.meta_fp)
        self.meta = read_metadata(self.meta_fp)

    def set_coassembly(self):
        """Create a metadata variable for the groups on which to co-assemble."""
        if self.coassembly:
            for name, coassembly in self.coassembly.items():
                self.get_pooling_groups(name, coassembly)
        else:
            self.pooling_groups = ['assembly_per_sample']
            self.meta['assembly_per_sample'] = self.meta.sample_name

    def get_pooling_groups(
            self,
            name: str,
            coassembly: dict
    ) -> None:
        """Get the column(s) corresponding to the factors groups
        identifying the samples to co-assemble.

        Parameters
        ----------
        name : str
            Name of the co-assembly.
        coassembly : dict
            Factors groups identifying the samples to co-assemble
        """
        cols = {}
        for col, factors in coassembly.items():
            if col not in self.meta.columns:
                raise IOError(
                    'Co-assembly variable "%s" not in %s' % (col, self.meta_fp))
            factors_flat = [str(f) for factor in factors for f in factor]
            if not set(factors_flat).issubset(set(self.meta[col])):
                raise IOError('Co-assembly factors for variable "%s" not'
                              ' in %s' % (col, self.meta_fp))
            d = {str(y): '_'.join(map(str, x)) for x in factors for y in x}
            cols[col] = [d[x] if x in d else np.nan for x in self.meta[col]]
        ser = pd.DataFrame(cols).fillna('')
        col = ser.apply(lambda x: '-'.join([str(i) for i in x if i]), axis=1)
        self.meta[name] = col.replace('', np.nan)
        self.pooling_groups.append(name)

    def get_fastq_paths(self) -> list:
        fastqs = []
        for fastq_dir in self.fastq_dirs:
            fastqs.extend(glob.glob(fastq_dir + '/*.fastq*'))
        return fastqs

    def get_fastq_samples(self):
        fastq_paths = self.get_fastq_paths()
        fastq = fill_fastq(fastq_paths, set(self.meta.sample_name))
        # keep only the `.fastq.gz` files (if `.fastq` files are also present)
        for sam, fastqs_ in fastq.items():
            if len([x for x in fastqs_ if '.gz' in x]):
                fqs = [x for x in fastqs_ if '.gz' in x]
            else:
                fqs = fastqs_
            self.fastq[sam] = fqs
            self.fastq_scratch[sam] = ['${SCRATCH_FOLDER}%s' % x for x in fqs]

    def set_fastq(self):
        """
        Check that fastq folder exists and that it contains fastq files.
        """
        if sum([not exists(x) for x in self.fastq_dirs]):
            raise IOError('Fastq folder(s) do not exist')
        self.get_fastq_samples()

    def get_r(self):
        self.r = {}
        for sam, r1_r2 in self.fastq.items():
            self.r[sam] = []
            if len(r1_r2) == 2:
                r1, r2 = r1_r2
                diffs = [i for i in range(len(r1)) if r1[i] != r2[i]]
                if len(diffs) > 1:
                    raise IOError('Too different fastq names for sample "%s":\n'
                                  '- %s\n- %s' % (sam, r1, r2))
                self.r[sam] = [r1[diffs[0] - 1:].split('.fastq')[0],
                               r2[diffs[0] - 1:].split('.fastq')[0]]

    def update_metadata(self):
        self.meta.set_index('sample_name', inplace=True)
        self.meta = self.meta.loc[list(self.fastq.keys())]
        self.meta.fillna('Unspecified', inplace=True)
        self.meta_fp = '%s_pipeline.tsv' % splitext(self.meta_fp)[0]
        self.meta.to_csv(self.meta_fp, sep='\t')

    def set_output(self):
        """Check if main output folder exists or create it."""
        if not isdir(self.output_dir):
            os.makedirs(self.output_dir)
        self.dir = os.path.abspath(self.output_dir)

    def parse_yamls(self):
        for arg in list(self.__dict__):
            if arg.endswith('_yml'):
                yaml = read_yaml(self.__dict__[arg])
                setattr(self, arg[:-4], yaml)
            if arg == 'pipeline_tsv':
                with open(self.__dict__[arg]) as f:
                    read = [['edit']] + [
                        x.strip().split() for x in f.readlines()
                        if x[0] != '#' and len(x.strip().split())]
                    setattr(self, arg[:-4], read)

    def get_default_params(self):
        """
        Get the run parameters based on the default
        values, that are possibly updated by the user.
        """
        if isfile('%s/run_params.yml' % RESOURCES):
            self.params = read_yaml('%s/run_params.yml' % RESOURCES)['default']



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
import pkg_resources
from os.path import isfile, splitext
from metagenomix._io_utils import io_update

# Keep line because read mapping alignments will be python scripts
# scripts = pkg_resources.resource_filename('metagenomix', 'resources/scripts')


def condition_ali_command(fastx: list, cmd: str, sam: str) -> str:
    """Add the conditional statements checking that the alignment
    already available was not aborted, in which case the alignment
    will be re-run.

    Parameters
    ----------
    fastx : list
        Path to the input fasta/fastq/fastq.gz files
    cmd : str
        Alignment command
    sam : str
        Path to the alignment file

    Returns
    -------
    cmd : str
        Alignment command potentially decorated with unix check to rerun
    """
    if fastx[-1].endswith('fastq.gz'):
        reads = 'zcat %s | tail -n 80 | grep "^@"' % fastx[-1]
    elif fastx[-1].endswith('fastq'):
        reads = 'tail -n 80 %s | grep "^@"' % fastx[-1]
    elif fastx[-1].endswith('fasta'):
        reads = 'tail -n 40 %s | grep "^>"' % fastx[-1]
    else:
        return cmd

    cmd = grep_sam_tail_cmd(cmd, reads, sam)
    return cmd


def grep_sam_tail_cmd(cmd, reads, sam) -> str:
    grep_cmd = '\nlast_reads=`%s`\n' % reads
    grep_cmd += 'GREPOK=0\n'
    grep_cmd += 'for i in $listVar\n'
    grep_cmd += 'do\n'
    grep_cmd += '    if grep -q "$i" %s\n' % sam
    grep_cmd += '        then\n'
    grep_cmd += '            GREPOK=1\n'
    grep_cmd += '            break\n'
    grep_cmd += '    fi\n'
    grep_cmd += 'done\n'
    grep_cmd += "if [ $GREPOK = 0 ]\n"
    grep_cmd += 'then\n'
    grep_cmd += '%s\n' % cmd
    grep_cmd += 'fi\n'
    return grep_cmd


def get_burst_cmd(fastx: str, db_path: str, params: dict, out: str) -> str:
    """Get the burst alignment command.

    Parameters
    ----------
    fastx : str
        Path to the input fasta file
    db_path : str
        Path to the burst database
    params : dict
        Run parameters.
    out : str
        Path to the output folder

    Returns
    -------
    cmd : str
        Alignment command
    """
    cmd = 'burst'
    cmd += ' -q %s' % fastx
    cmd += ' -r %s.edx' % db_path
    cmd += ' -a %s.acx' % db_path
    cmd += ' -sa'
    cmd += ' -fr'
    cmd += ' -i 0.98'
    cmd += ' -t %s' % params['cpus']
    cmd += ' -o %s' % out
    return cmd


def get_bowtie2_cmd(self, fastx: list, db_path: str, db_out: str) -> tuple:
    """Get the bowtie2 alignment command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters for humann
    fastx : list
        Path to the input fasta/fastq/fastq.gz files
    db_path : str
        Path to the bowtie2 database
    db_out: str
        Path to the output folder

    Returns
    -------
    cmd : str
        Alignment command
    sam : str
        Alignment .sam file
    """
    cmd = 'bowtie2'
    cmd += ' -p %s' % self.soft.params['cpus']
    cmd += ' -x %s' % db_path
    sam = '%s/alignment.bowtie2' % db_out
    if len(fastx) == 2 and self.soft.params['pairing'] == 'paired':
        cmd += ' -1 %s' % fastx[0]
        cmd += ' -2 %s' % fastx[1]
        if not self.soft.params['discordant']:
            cmd += ' --no-discordant'
            sam += '.nodiscordant'
    else:
        cmd += ' -U %s' % ','.join(fastx)
    sam += '.sam'
    cmd += ' -S %s' % sam
    cmd += ' --seed 12345'
    cmd += ' --very-sensitive'
    cmd += ' -k %s' % self.soft.params['k']
    cmd += ' --np %s' % self.soft.params['np']
    cmd += ' --mp "%s"' % self.soft.params['mp']
    cmd += ' --rdg "%s"' % self.soft.params['rdg']
    cmd += ' --rfg "%s"' % self.soft.params['rfg']
    cmd += ' --score-min "%s"' % self.soft.params['score-min']
    cmd += ' --met-file %s.met' % splitext(sam)[0]
    cmd += ' --met 240'
    cmd += ' --no-unal'
    return cmd, sam


def get_alignment_cmd(fastx: list, cmd: str, ali: str) -> str:
    """Get the alignment command with ot without unix checks to decide
    whether the alignment (if exists) was aligned to completion, and not
    aborted (e.g., for lack of memory reasons).

    Parameters
    ----------
    fastx : list
        Path to the input fasta/fastq/fastq.gz files
    cmd : str
        Alignment command
    ali : str
        Path to the alignment file

    Returns
    -------
    cmd : str
        Alignment command potentially decorated with unix check to rerun
    """
    if not isfile(ali):
        return cmd
    else:
        return condition_ali_command(fastx, cmd, ali)


def bowtie2(self) -> None:
    """Get the full command lines to perform sequence alignment using
     bowtie2 and potentially, re-run if the output was existing but
     possibly aborted.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
    """
    out = self.dir + '/' + self.sam
    self.outputs['outs'] = dict()
    fastx = self.inputs[self.sam]
    for db, db_path in self.soft.params['databases'].items():
        db_out = '%s/%s/%s' % (out, db, self.soft.params['pairing'])
        cmd, sam = get_bowtie2_cmd(self, fastx, db_path, db_out)
        cmd = get_alignment_cmd(fastx, cmd, sam)
        key = (db, self.soft.params['pairing'])
        self.outputs['outs'][key] = sam
        self.outputs['cmds'].append(cmd)
        self.outputs['dirs'].append(db_out)
        io_update(self, i_f=fastx, o_d=db_out)


def check_bowtie_k_np(soft, params, defaults, let_go):
    for opt in ['k', 'np']:
        if opt in params and not str(params[opt]).isdigit():
            sys.exit('[bowtie2] "%s" option invalid' % opt)
        else:
            soft.params[opt] = defaults[opt]
        let_go.append(opt)


def check_bowtie_mp_rdg_rfg(soft, params, defaults, let_go):
    for opt in ['mp', 'rdg', 'rfg']:
        if opt in params:
            if len([x.isdigit() for x in str(params[opt]).split(',')]) != 2:
                sys.exit('[bowtie2] "%s" option invalid' % opt)
        else:
            soft.params[opt] = defaults[opt]
        let_go.append(opt)


def check_bowtie_score_min(soft, params, defaults, let_go):
    if 'score-min' in params:
        s = params['score-min'].split(',')
        if len(s) != 3 or s[0] not in 'CLSG':
            sys.exit('[bowtie2] "score-min" option invalid')
        else:
            for r in [1, 2]:
                try:
                    float(s[r])
                except ValueError:
                    sys.exit('[bowtie2] "score-min" option invalid')
    else:
        soft.params['score-min'] = defaults['score-min']
    let_go.append('score-min')


def flash(self):
    """Create command lines for FLASh

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .config
            Configurations
    """
    min_overlap = self.soft.params['min_overlap']
    max_overlap = self.soft.params['max_overlap']
    mismatch = self.soft.params['mismatch']
    out = '%s/m%s_M%s_e%s/%s' % (
        self.dir, min_overlap, max_overlap, mismatch, self.sam)
    rad = out + '/' + self.sam
    ext_fq = '%s.extendedFrags.fastq' % rad
    ext = ext_fq.replace('.fastq', '.fasta')
    nc1_fq = '%s.notCombined_1.fastq' % rad
    nc1 = nc1_fq.replace('.fastq', '.fasta')
    nc2_fq = '%s.notCombined_2.fastq' % rad
    nc2 = nc2_fq.replace('.fastq', '.fasta')
    out_fps = [ext, nc1, nc2]
    self.outputs['dirs'].append(out)
    self.outputs['outs'].extend(out_fps)
    if self.config.force or len([not isfile(x) for x in out_fps]):
        cmd = 'flash %s' % ' '.join(self.inputs[self.sam])
        cmd += ' -m %s' % min_overlap
        cmd += ' -M %s' % max_overlap
        cmd += ' -x %s' % mismatch
        cmd += ' -d %s' % out
        cmd += ' -o %s' % self.sam
        cmd += ' -t %s\n' % self.soft.params['cpus']
        cmd += 'seqtk seq -A %s > %s\n' % (ext_fq, ext)
        cmd += 'seqtk seq -A %s > %s\n' % (nc1_fq, nc1)
        cmd += 'seqtk seq -A %s > %s\n' % (nc2_fq, nc2)
        self.outputs['cmds'].append(cmd)
        io_update(self, i_f=self.inputs[self.sam], o_f=out_fps)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
import gzip
from os.path import basename, isfile
from metagenomix._io_utils import io_update


def get_cat_zcat(fastq_fp: str) -> str:
    """Get the command `cat` or `gzcat` to run for fastq editing.

    Parameters
    ----------
    fastq_fp : str
        Fastq file path.

    Returns
    -------
    cat : str
        The command to use (`cat` or `zcat`).
    """
    cat = "cat"
    if fastq_fp.endswith('.fastq.gz'):
        cat = "zcat"
    return cat


def edit_fastq_cmd(fastq_fp: str, num: int, source: str) -> str:
    """Get the unix command to run on each fastq file that
    needs editing, depending on the source of fastq file.

    Parameters
    ----------
    fastq_fp : str
        Fastq file path.
    num : int
        Can be 1 or 2 for the forward and reverse read, respectively.
    source : str
        Source of the fastq that defines format and editing type.

    Returns
    -------
    cmd : str
        Unix command to run to edit a fastq file.
    """
    cat = get_cat_zcat(fastq_fp)
    cmd = '%s %s | ' % (cat, fastq_fp)
    cmd += "awk '{ if (NR%s4==1) " % '%'
    if source == 'illumina':
        cmd += "{ print $1\"/%s\" } " % str(num)
    elif source == 'ebi':
        cmd += "{ gsub(/ .*/,\"/%s\",$0); print } " % num
    cmd += "else if (NR%s2 == 1) { print \"+\" } " % '%'
    cmd += "else { print } }'"
    if fastq_fp.endswith('.gz'):
        cmd += " | gzip"
    cmd += " > %s_renamed\n" % fastq_fp
    cmd += "mv %s_renamed %s" % (fastq_fp, fastq_fp)
    if fastq_fp.endswith('.fastq'):
        cmd += "\n"
    else:
        cmd += ".gz\n"
    return cmd


def get_fastq_header(fastq_path: str) -> str:
    """Get the first line of the fastq file.

    Parameters
    ----------
    fastq_path : str
        Path to a fastq.gz or fastq file.

    Returns
    -------
    fastq_line : str
        First line of the fastq file.
    """
    fastq_line = ''
    if fastq_path.endswith('fastq.gz'):
        with gzip.open(fastq_path) as f:
            for line in f:
                fastq_line = line.decode().strip()
                break
    elif fastq_path.endswith('fastq'):
        with open(fastq_path) as f:
            for line in f:
                fastq_line = line.strip()
                break
    return fastq_line


def get_edit_cmd(self, num: int) -> None:
    """Get the unix command to run on each fastq file that needs editing.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .config
            Configurations
    num : int
        Can be 1 or 2 for the forward and reverse read, respectively.
    """
    fastqs_fps = self.inputs[self.sam]
    fastq_fp = fastqs_fps[num - 1]
    line = get_fastq_header(self.config.fastq[self.sam][num - 1])
    line_split = line.split()
    if not line_split[0].endswith('/%s' % num):
        if len(line_split) > 1:
            if line_split[1].startswith('%s:N:' % num):
                cmd = edit_fastq_cmd(fastq_fp, num, 'illumina')
            elif line_split[0].endswith('.%s' % num):
                cmd = edit_fastq_cmd(fastq_fp, num, 'ebi')
            else:
                sys.exit('Fastq file headers not editable:\n%s' % line)
            self.outputs['cmds'].append(cmd)
    io_update(self, i_f=fastqs_fps, o_f=fastqs_fps)


def edit(self) -> None:
    """Create command lines for editing the fasta if need be so that
    the fastq header all contaim the trailing "/1" and "/2" that allow
    identifying paired reads.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .config
            Configurations
    """
    for r in [1, 2]:
        if len(self.config.fastq[self.sam]) < r:
            continue
        get_edit_cmd(self, r)


def count_cmd(self, idx: int, input_fp: str, out: str) -> str:
    """Get the command to count the reads in a fasta, fastq or fastq.gz file.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .soft.params
            Parameters
    idx : int
        Unique, incremental numeric identifier.
    input_fp : str
        Path to the input file name.
    out : str
        Path to the output file name.

    Returns
    -------
    cmd : str
        Command to count reads and output the counts.
    """
    if input_fp.endswith('fastq.gz'):
        cmd = "n%s=`%s %s | wc -l | " % (idx, self.soft.params['cat'], input_fp)
        cmd += "sed 's/ //g' | awk '{x=$1/4; print x}'`\n"
        # cmd += "cut -d ' ' -f 1 | awk '{x=$1/4; print x}'`"
    elif input_fp.endswith('fasta'):
        cmd = "n%s=`wc -l %s | " % (idx, input_fp)
        cmd += "sed 's/ //g' | awk '{x=$1/2; print x}'`\n"
        # cmd += "cut -d ' ' -f 1 | awk '{x=$1/2; print x}'`"
    elif input_fp.endswith('fastq'):
        cmd = "n%s=`wc -l %s | " % (idx, input_fp)
        cmd += "sed 's/ //g' | awk '{x=$1/4; print x}'`\n"
        # cmd += "cut -d ' ' -f 1 | awk '{x=$1/4; print x}'`"
    else:
        raise IOError("Input sequence file invalid %s" % input_fp)
    if idx:
        cmd += 'echo "%s,2,${n%s}" >> %s\n' % (self.sam, idx, out)
    else:
        cmd += 'echo "%s,1,${n%s}" > %s\n' % (self.sam, idx, out)
    return cmd


def count(self) -> None:
    """Create command lines for counting the reads.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    out = '%s/%s_read_count.tsv' % (self.dir, self.sam)
    self.outputs['outs'].append(out)
    if self.config.force or not isfile(out):
        inputs = self.inputs[self.sam]
        for idx, input_path in enumerate(inputs):
            cmd = count_cmd(self, idx, input_path, out)
            self.outputs['cmds'].append(cmd)
        io_update(self, i_f=inputs, o_f=out)


def fastqc(self) -> None:
    out_dir = '%s/%s' % (self.dir, self.sam)
    ins = self.inputs[self.sam]
    outs = ['%s_fastqc.html' % x.rsplit('.fastq', 1)[0] for x in ins]
    self.outputs['outs'].append(out_dir)
    if self.config.force or [not isfile(x) for x in outs]:
        cmd = 'fastqc %s -o %s' % (' '.join(ins), out_dir)
        self.outputs['dirs'].append(out_dir)
        self.outputs['cmds'].append(cmd)
        io_update(self, i_f=ins, o_d=out_dir)


def fastp(self) -> None:
    fastqs = self.inputs[self.sam]
    out_dir = '%s/%s' % (self.dir, self.sam)
    out_json = '%s/%s.json' % (out_dir, self.sam)
    out_html = '%s/%s.html' % (out_dir, self.sam)
    outs = []
    self.outputs['outs'].append(out_dir)
    cmd = 'fastp'
    for fdx, fastq in enumerate(fastqs):
        if fdx:
            i, o = 'I', 'O'
            out = '%s/%s_2.fastq' % (out_dir, self.sam)
        else:
            i, o = 'i', 'o'
            out = '%s/%s_1.fastq' % (out_dir, self.sam)
        if fastq.endswith('.gz'):
            out += '.gz'
        outs.append(isfile(out))
        cmd += '  -%s %s -%s %s' % (i, out, o, out)
    cmd += ' --verbose'
    if self.soft.params['trim_poly_g']:
        cmd += ' --trim_poly_g'
    cmd += ' --json=%s' % out_json
    cmd += ' --html=%s' % out_html
    cmd += ' -w %s' % self.soft.params['cpus']
    cmd += ' --report_title="fastp report %s"' % self.sam
    cmd += ' --average_qual=%s' % self.soft.params['average_qual']
    if self.config.force or not sum(outs):
        self.outputs['dirs'].append(out_dir)
        self.outputs['cmds'].append(cmd)
        io_update(self, i_f=fastqs, o_d=out_dir)


def cutadapt(self) -> None:
    r1_o = '%s/%s.R1.fastq.gz' % (self.dir, self.sam)
    r2_o = '%s/%s.R2.fastq.gz' % (self.dir, self.sam)
    cmd = 'cutadapt'
    cmd += ' -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCA'
    cmd += ' -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGT'
    cmd += ' -m 10'
    cmd += ' -o %s' % r1_o
    cmd += ' -p %s ' % r2_o
    cmd += ' '.join(self.inputs[self.sam])
    outs = [r1_o, r2_o]
    self.outputs['outs'].extend(outs)
    if self.config.force or not isfile(r1_o) or not isfile(r2_o):
        self.outputs['cmds'] = list([cmd])
        io_update(self, i_f=self.inputs[self.sam], o_f=outs)


def atropos(self):
    inputs = self.inputs[self.sam]
    if len(inputs) == 1:
        out_1 = '%s/%s/%s.fastq' % (self.dir, self.sam, self.sam)
    else:
        out_1 = '%s/%s/%s_R1.fastq' % (self.dir, self.sam, self.sam)
    out_2 = '%s/%s/%s_R2.fastq' % (self.dir, self.sam, self.sam)
    log = '%s/%s/%s.log' % (self.dir, self.sam, self.sam)
    if inputs[0].endswith('.gz'):
        out_1 += '%s.gz'
        out_2 += '%s.gz'
    cmd = 'atropos trim'
    for a in self.soft.params['a']:
        cmd += ' -a %s' % a
    for A in self.soft.params['A']:
        cmd += ' -A %s' % A
    cmd += ' -q %s' % self.soft.params['q']
    cmd += ' --minimum-length %s' % self.soft.params['minimum_length']
    cmd += ' --nextseq-trim %s' % self.soft.params['nextseq_trim']
    cmd += ' --pair-filter %s' % self.soft.params['pair_filter']
    cmd += ' --overlap %s' % self.soft.params['overlap']
    cmd += ' --max-reads %s' % self.soft.params['max_reads']
    cmd += ' --indel-cost %s' % self.soft.params['indel_cost']
    cmd += ' --error-rate %s' % self.soft.params['error_rate']
    cmd += ' --quality-cutoff %s' % self.soft.params['quality_cutoff']
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --report-file %s' % log
    cmd += ' --report-formats %s' % self.soft.params['report_formats']
    if len(inputs) == 1:
        cmd += ' -se %s' % inputs[0]
        cmd += ' -o %s' % out_1
    else:
        cmd += ' -pe1 %s' % inputs[0]
        cmd += ' -pe2 %s' % inputs[1]
        cmd += ' -o %s' % out_1
        cmd += ' -p %s' % out_2
    if self.config.force or not isfile(out_1) or not isfile(out_2):
        self.outputs['cmds'].append(cmd)
        io_update(self, i_f=inputs, o_f=[out_1, out_2])
    self.outputs['outs'].extend([out_1, out_2])


def kneaddata(self):
    inputs = self.inputs[self.sam]
    out_dir = '%s/%s' % (self.dir, self.sam)
    log = '%s/%s.log' % (out_dir, self.sam)
    cmd = 'kneaddata'
    for inp in inputs:
        cmd += ' --input %s' % inp
    for database in self.soft.params['databases']:
        cmd += ' --reference-db %s' % database
    cmd += ' --output %s' % out_dir
    # cmd += ' --output-prefix %s' % self.sam
    cpus = self.soft.params['cpus']
    cmd += ' --processes %s' % cpus
    cmd += ' --max-memory %s%s' % (self.soft.params['mem_num'],
                                   self.soft.params['mem_dim'][0])
    if self.soft.params['trimmomatic']:
        cmd += ' --trimmomatic %s' % self.soft.params['trimmomatic']
    if self.soft.params['bowtie2']:
        cmd += ' --bowtie2 %s' % self.soft.params['bowtie2']
    cmd += ' --bowtie2-options="--very-sensitive-local -p %s"' % cpus
    cmd += ' --remove-intermediate-output'
    cmd += ' --log %s' % log
    self.outputs['cmds'].append(cmd)
    # write these additional output files manipulations commands
    rad = basename(inputs[0]).replace('.fastq.gz', '')
    for typ in ['paired', 'unmatched']:
        for r in ['1', '2']:
            out = '%s/%s_kneaddata_%s_%s.fastq' % (out_dir, rad, typ, r)
            renamed = out.replace('.R1', '')
            self.outputs['outs'].append(renamed)
            if out == renamed:
                continue
            self.outputs['cmds'].append('mv %s %s' % (out, renamed))

    if self.soft.params['purge']:
        tar_cmd = 'rm %s/%s*' % (out_dir, rad)
    else:
        tar_cmd = 'tar cpfz %s/%s_kneadfiles.tar.gz %s/%s* --remove-files' % (
            out_dir, self.sam, out_dir, rad)

    self.outputs['cmds'].append(tar_cmd)
    io_update(self, i_f=inputs, o_d=out_dir)


def filtering(self):
    outputs = self.inputs[self.sam]
    io_update(self, i_f=self.inputs[self.sam])
    cmds = ''
    databases = self.soft.params['databases']
    for ddx, db_path in enumerate(databases):
        inputs = list(outputs)
        outputs = []
        db = db_path.split('/')[-1]
        if len(inputs) == 1:
            out_1 = '%s/%s_%s/%s.fastq' % (self.dir, ddx, db, self.sam)
        else:
            out_1 = '%s/%s_%s/%s_R1.fastq' % (self.dir, ddx, db, self.sam)
        out_2 = '%s/%s_%s/%s_R2.fastq' % (self.dir, ddx, db, self.sam)
        if inputs[0].endswith('.gz'):
            out_1 += '%s.gz'
            out_2 += '%s.gz'

        cmd = '\nbowtie2'
        cmd += ' -p %s' % self.soft.params['cpus']
        cmd += ' -x %s' % db_path
        cmd += ' --very-sensitive'
        if len(inputs) == 1:
            cmd += ' -U %s' % inputs[0]
        else:
            cmd += ' -1 %s' % inputs[0]
            cmd += ' -2 %s' % inputs[1]
        cmd += ' | samtools view -f 12 -F 256'
        cmd += ' | samtools sort -@ %s -n' % self.soft.params['cpus']
        cmd += ' | samtools view -bS'
        cmd += ' | bedtools bamtofastq -i -'
        cmd += ' -fq %s' % out_1.replace('fastq.gz', 'fastq')
        outputs.append(out_1.replace('fastq.gz', 'fastq'))
        if len(inputs) > 1:
            cmd += ' -fq2 %s' % out_2.replace('fastq.gz', 'fastq')
            outputs.append(out_2.replace('fastq.gz', 'fastq'))
        if ddx:
            cmd += '\nrm %s\n' % ' '.join(inputs)
        if self.config.force or not isfile(out_1):
            cmds += cmd
    if outputs[0].endswith('.fastq'):
        for output in outputs:
            cmds += '\ngzip %s' % output
    if cmds:
        self.outputs['cmds'].append(cmds)
    self.outputs['outs'].extend(outputs)
    io_update(self, o_f=outputs)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import subprocess
from os.path import isfile
from typing import TextIO


def run_xpbs(out_sh: str, out_pbs: str, job_name: str,
             qiime_env: str, time: str, n_nodes: str,
             n_procs: str, mem_num: str, mem_dim: str, chmod: str,
             written: int, single: str, o: TextIO = None,
             noloc: bool = True, slurm: bool = False, jobs: bool = True,
             tmp: str = None) -> None:
    """
    Run the Xpbs script assorted with print or writing in higher-level command.
    """
    if written:
        if os.getcwd().startswith('/panfs'):
            out_sh_lines = open(out_sh).readlines()
            with open(out_sh, 'w') as sh:
                if not jobs:
                    sh.write('conda activate %s\n' % qiime_env)
                for out_sh_line in out_sh_lines:
                    sh.write(out_sh_line.replace(os.getcwd(), ''))
        if jobs:
            xpbs_call(out_sh, out_pbs, job_name, qiime_env,
                      time, n_nodes, n_procs, mem_num,
                      mem_dim, chmod, slurm, noloc, tmp)
        if single:
            if slurm:
                launcher = 'sbatch'
            else:
                launcher = 'qsub'
            if os.getcwd().startswith('/panfs'):
                out_pbs = out_pbs.replace(os.getcwd(), '')
            if not o:
                print(single)
                if jobs:
                    print('[TO RUN]', launcher, out_pbs)
                else:
                    print('[TO RUN] sh', out_sh)
            else:
                if jobs:
                    o.write('%s %s\n' % (launcher, out_pbs))
                else:
                    o.write('sh %s\n' % out_sh)
    else:
        os.remove(out_sh)
        if isfile(out_pbs):
            os.remove(out_pbs)


def xpbs_call(out_sh: str, out_pbs: str, prjct_nm: str,
              qiime_env: str, time: str, n_nodes: str,
              n_procs: str, mem_num: str, mem_dim: str,
              chmod: str, slurm: bool, noloc: bool,
              tmp: str = None) -> None:
    """
    Call the subprocess to run Xpbs on the current bash script.
    """
    cmd = [
        'Xpbs',
        '-i', out_sh,
        '-j', prjct_nm,
        '-o', out_pbs,
        '-e', qiime_env,
        '-t', time,
        '-n', n_nodes,
        '-p', n_procs,
        '-M', mem_num, mem_dim,
        '-c', chmod,
        '--noq'
    ]
    if tmp:
        cmd.extend(['-T', tmp])
    if not noloc:
        cmd.append('--no-loc')
    if slurm:
        cmd.append('--slurm')
    subprocess.call(cmd)

    if os.getcwd().startswith('/panfs'):
        out_pbs_lines = open(out_pbs).readlines()
        with open(out_pbs, 'w') as pbs:
            for out_pbs_line in out_pbs_lines:
                pbs.write(out_pbs_line.replace(os.getcwd(), ''))


def print_message(message: str, sh_pbs: str, to_run: str, jobs: bool) -> None:
    if message:
        print(message)
    if os.getcwd().startswith('/panfs'):
        to_run = to_run.replace(os.getcwd(), '')
    if jobs:
        print('[TO RUN]', sh_pbs, to_run)
    else:
        print('[TO RUN]', 'sh', to_run.replace('.pbs', '.sh'))



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import sys
import pkg_resources
from os.path import basename, isdir, isfile, splitext

from metagenomix._io_utils import check_min_lines_count, io_update
from metagenomix.tools.alignment import get_alignment_cmd

RESOURCES = pkg_resources.resource_filename('metagenomix', 'resources')


def shogun_append_cmd(self, cmds: list, tab: str):
    """Add or not the current SHOGUN command to the list of commands to run
    depending on whether the file already exists or exists but is only a header.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    cmds : list
        List of current SHOGUN command lines
    tab : str
        Path to the output table of the current command
    """
    if self.config.force:
        self.outputs['cmds'].extend(cmds)
    elif not isfile(tab) or not check_min_lines_count(tab):
        cmd = 'file="%s"\n' % tab
        cmd += 'if [ -f "$file" ]\n'
        cmd += 'then\n'
        cmd += "FILESIZE=`ls -l $file | cut -d ' ' -f5`\n"
        cmd += "if [[ $FILESIZE -lt 1000 ]]\n"
        cmd += 'then\n'
        cmd += '\n'.join(cmds)
        cmd += 'fi\n'
        self.outputs['cmds'].append(cmd)


def shogun_redistribute(self, db: str, aligner: str, tax_norm: str,
                        db_path: str) -> None:
    """Get the SHOGUN command to redistribute a taxonomic profile
    at various taxonomic levels.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    db : str
        Database name
    aligner : str
        Name of the aligner
    tax_norm : str
        Path to the input taxonomic table file
    db_path : str
        Path to the taxonomic database
    """
    redist_cmds = []
    for level in ['phylum', 'genus', 'strain']:
        redist = '%s.redist.%s.tsv' % (splitext(tax_norm)[0], level)
        cmd = 'shogun redistribute'
        cmd += ' -i %s' % tax_norm
        cmd += ' -d %s' % db_path
        cmd += ' -l %s' % level
        cmd += ' -o %s' % redist
        redist_cmds.append(cmd)
        self.outputs['outs'].setdefault((db, aligner), []).append(redist)
        io_update(self, o_f=redist)
    redist_out = '%s.redist.strain.tsv' % splitext(tax_norm)[0]
    shogun_append_cmd(self, redist_cmds, redist_out)


def shogun_assign_taxonomy(aligner: str, ali: str, tax: str, db_path: str,
                           sub_db: str) -> str:
    """Get the assign taxonomy command.

    Parameters
    ----------
    aligner : str
        Name of the aligner
    ali : str
        Path to the alignment
    tax : str
        Path to the output table
    db_path : str
        Path to the database
    sub_db : str
        Suffix to the database

    Returns
    -------
    cmd : str
        assign taxonomy command
    """
    cmd = 'shogun assign_taxonomy'
    cmd += ' -a %s' % aligner
    cmd += ' -i %s' % ali
    cmd += ' -d %s%s' % (db_path, sub_db)
    cmd += ' -o %s\n' % tax
    return cmd


def shogun_normalize(input_fp: str, output_fp: str) -> str:
    """Get the SHOGUN command to normalize table.

    Parameters
    ----------
    input_fp : str
        Path to the input table to normalize.
    output_fp : str
        Path to the output normalized table.

    Returns
    -------
    cmd : str
        Command line for table normalization.
    """
    cmd = 'shogun normalize -i %s -o %s\n' % (input_fp, output_fp)
    return cmd


def get_ali_cmd(self, fasta: str, out: str, db_path: str, aligner: str) -> str:
    """Get the SHOGUN alignment command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
        .databases
            All databases class instance
    fasta : str
        Path to the input fasta file
    db_path : str
        Path to the shogun database
    out : str
        Path to the output folder
    aligner : str
        Name of the aligner

    Returns
    -------
    cmd : str
        Alignment command using SHOGUN aligner
    """
    cmd = 'shogun align -a %s' % aligner
    cmd += ' -i %s' % fasta
    cmd += ' -d %s' % db_path
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' -o %s' % out
    return cmd


def get_alignment_basename(aligner):
    if aligner == 'bowtie2':
        ali_base = 'alignment.bowtie2.sam'
    elif aligner == 'burst':
        ali_base = 'alignment.burst.b6'
    else:
        raise ValueError('No output basename for aligner "%s"' % aligner)
    return ali_base


def get_orients(inputs: list):
    if len(inputs) == 1:
        orients = ['']
    elif len(inputs) == 2:
        orients = ['1', '2']
    elif len(inputs) == 3:
        orients = ['', '1', '2']
    else:
        raise IOError('Input to shogun must be 1, 2 or 3 fasta files')
    return orients


def get_combine_cmd(self, out_dir: str, combine_cmds: list) -> list:
    """

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
    out_dir : str
        Path to pipeline output folder for SHOGUN
    combine_cmds : list
        Commands to turn .fastq or .fastq.gz files into .fasta files

    Returns
    -------
    fastas : list
        Paths to fasta files to combine for SHOGUN
    """
    fastas = []
    # get the fastq versions of input file
    orients = get_orients(self.inputs[self.sam])
    for pdx, path_ in enumerate(sorted(self.inputs[self.sam])):
        io_update(self, i_f=path_)
        path = path_
        if path_.endswith('fastq') or path_.endswith('fastq.gz'):
            # replace non-fasta by fasta extensions
            path = path_.replace('.fastq', '.fasta').replace('.gz', '')
            # Prepare the extraction command and collect it
            if isfile(path_) and not isfile(path):
                to_fasta_cmd = 'seqtk seq -A %s > %s' % (path_, path)
                combine_cmds.append(to_fasta_cmd)

        path_out = '%s/%s' % (out_dir, basename(path))
        edit_fasta = 'python3 %s/scripts/fasta4shogun.py -i %s -o %s -s %s' % (
            RESOURCES, path, path_out, self.sam)
        orient = orients[pdx]
        if orient:
            edit_fasta += ' -r %s' % orient
        combine_cmds.append(edit_fasta)
        fastas.append(path_out)

    return fastas


def combine_inputs(self, out_dir: str, combine_cmds: list) -> str:
    """Combine the fastq/a of the sample to run shogun.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
    out_dir : str
        Path to pipeline output folder for SHOGUN
    combine_cmds : list
        Commands to turn .fastq or .fastq.gz files into .fasta files

    Returns
    -------
    fasta : str
        Path to the combined sequences fasta file
    """
    fastas = get_combine_cmd(self, out_dir, combine_cmds)
    fasta = '%s/combined.fasta' % out_dir
    combine_cmds.extend([
        'cat %s > %s' % (' '.join(fastas), fasta), 'rm %s' % ' '.join(fastas)])
    return fasta


def align_cmds(self, fasta: str, out: str, db: str, aligner: str,
               ali_cmds: list) -> str:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
    fasta : str
        Path to the combined sequences fasta file
    out : str
        Path to pipeline output folder for SHOGUN
    db : str
        Database name
    aligner : str
    ali_cmds : list

    Returns
    -------
    ali : str
        Alignment output file
    """
    out_dir = '%s/%s/%s' % (out, db, aligner)
    ali = '%s/%s' % (out_dir, get_alignment_basename(aligner))

    db_path = '%s/shogun' % self.databases.paths[db]
    cmd = get_ali_cmd(self, fasta, out_dir, db_path, aligner)
    cmd = get_alignment_cmd([fasta], cmd, ali)
    ali_cmds.append(cmd)
    if cmd.startswith('shogun'):
        io_update(self, i_f=ali)

    self.outputs['dirs'].append(out_dir)
    self.outputs['outs'].setdefault((db, aligner), []).append(ali)
    io_update(self, o_d=out_dir)

    return ali


def format_sam(sam_: str, ali_cmds: list, sample: str) -> str:
    sam = '%s_formatted.sam' % splitext(sam_)[0]
    cmd = '%s/scripts/sam4shogun.py -i %s -o %s -s %s' % (
        RESOURCES, sam_, sam, sample)
    ali_cmds.append(cmd)
    return sam


def get_dir(out_dir: str, db: str, aligner: str) -> str:
    """

    Parameters
    ----------
    out_dir : str
        Path to pipeline output folder for SHOGUN
    db : str
        Database name
    aligner : str
        Name of the aligner

    Returns
    -------
    out : str
        Path to pipeline output folder for SHOGUN
    """
    out = '%s/%s' % (out_dir, db)
    if aligner:
        # i.e., if the previous software is not bowtie2
        out += '/%s' % aligner
    return out


def shogun_assign_normalize(self, tab: str, aligner: str, ali: str,
                            db_path: str, sub_db: str):
    """Get the assignment and the normalization SHOGUN commands for
    the taxonomy or functional classifications.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    tab : str
        Path to the output table
    aligner : str
        Name of the aligner
    ali : str
        Path to the alignment
    db_path : str
        Path to the database
    sub_db : str
        Suffix to the database
    """
    cmd = ''
    tab_norm = '%s_norm.tsv' % splitext(tab)[0]
    if self.config.force or not isfile(tab):
        cmd += shogun_assign_taxonomy(aligner, ali, tab, db_path, sub_db)
    if self.config.force or not isfile(tab_norm):
        cmd += shogun_normalize(tab, tab_norm)
    if cmd:
        shogun_append_cmd(self, list([cmd]), tab_norm)


def get_paths(self, out: str, aligner: str, db: str, tax_fun: str) -> tuple:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
    out : str
        Path to pipeline output folder for SHOGUN
    aligner : str
        Name of the aligner
    db : str
        Database name
    tax_fun : str
        "taxonomy" or "function"

    Returns
    -------
    tax : str
        Path to the output table
    tax_norm : str
        Path to the output table (relative abundances)
    """
    out_dir = get_dir(out, db, aligner)
    self.outputs['dirs'].append(out_dir)

    tab = '%s/%s.tsv' % (out_dir, tax_fun)
    tab_norm = '%s/%s_norm.tsv' % (out_dir, tax_fun)
    self.outputs['outs'].setdefault((db, aligner), []).extend([tab, tab_norm])
    io_update(self, i_d=out_dir, o_f=[tab, tab_norm])
    return tab, tab_norm


def shogun_coverage_cmd(self, ali: str, db_path: str, level: str,
                        cov_tab: str) -> None:
    """Get the SHOGUN command that calculates the coverage per taxon.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    ali : str
        Path to the alignment
    db_path : str
        Path to the taxonomic database
    level : str
        Taxonomic level for coverage calculation
    cov_tab : str
        Path to the output coverage table file
    """
    cmd = 'shogun coverage'
    cmd += ' -i %s' % ali
    cmd += ' -d %s' % db_path
    cmd += ' -l %s' % level
    cmd += ' -o %s' % cov_tab
    shogun_append_cmd(self, list([cmd]), cov_tab)


def shogun_coverage(self, out: str, aligner: str, ali: str, db: str) -> None:
    db_path = '%s/shogun' % self.databases.paths[db]
    if aligner == 'burst':
        cov_tab = '%s/coverage.tsv' % out
        self.outputs['outs'].setdefault((db, aligner), []).append(cov_tab)
        if self.config.force or not isfile(cov_tab):
            shogun_coverage_cmd(self, ali, db_path, 'strain', cov_tab)
            io_update(self, o_f=cov_tab)


def shogun_taxonomy(self, out: str, aligner: str, ali: str, db: str) -> str:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    out : str
        Path to pipeline output folder for SHOGUN
    aligner : str
        Name of the aligner
    ali : str
        Path to the alignment
    db : str
        Database name

    Returns
    -------
    norm : str
        Path to the normalized output table
    """
    db_path = '%s/shogun' % self.databases.paths[db]
    tab, norm = get_paths(self, out, aligner, db, 'taxonomy')
    shogun_assign_normalize(self, tab, aligner, ali, db_path, '')
    shogun_redistribute(self, db, aligner, norm, db_path)
    return norm


def shogun_function(self, out: str, aligner: str, ali: str, db: str,
                    norm: str) -> None:
    db_path = '%s/shogun' % self.databases.paths[db]
    for sub in ['kegg', 'refseq', 'uniprot']:
        fun, f_norm = get_paths(self, out, aligner, db, 'function-%s' % sub)
        shogun_assign_normalize(self, fun, aligner, ali, db_path,
                                '/functions-%s' % sub)
        shogun_redistribute(self, db, aligner, f_norm, db_path)

    out_dir = out + '/functional'
    for level in ['genus', 'species']:
        shogun_functional(self, norm, fun, out_dir, level)


def shogun_functional(self, norm: str, db_path: str, out_dir: str,
                      level: str) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    norm : str
        Path to the input table file
    db_path : str
        Path to the functional database
    out_dir : str
        Path to the output functional table
    level : str
        Taxonomic level
    """
    cmd = 'shogun functional'
    cmd += ' -i %s' % norm
    cmd += ' -o %s' % out_dir
    cmd += ' -d %s' % db_path
    cmd += ' -l %s' % level
    base = splitext(basename(norm))[0]
    out = '%s/%s.%s.normalized.txt' % (out_dir, base, level)
    shogun_append_cmd(self, list([cmd]), out)


def shogun(self) -> None:
    """Get the SHOGUN commands that consist of the classifications using the
    databases built for shogun and the alignment in shogun, or not as this
    may be run after bowtie2, in which case the classifications are performed
    on the alignment obtained before, provided that it was done on a database
    formatted for SHOGUN.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .databases
            All databases class instance
        .config
            Configurations
    """
    self.outputs['outs'] = dict({})
    combine_cmds, ali_cmds = [], []

    out = self.dir + '/' + self.sam
    self.outputs['dirs'].append(out)
    io_update(self, o_d=out)

    if self.soft.prev == 'bowtie2':
        for (db, pairing), sam_ in self.inputs[self.sam].items():
            ali = format_sam(sam_, ali_cmds, self.sam)
            shogun_taxonomy(self, out, '', ali, db)
    elif self.soft.params['databases']:
        fasta = combine_inputs(self, out, combine_cmds)
        for db, aligners in self.soft.params['databases'].items():
            for aligner in aligners:
                ali = align_cmds(self, fasta, out, db, aligner, ali_cmds)
                norm = shogun_taxonomy(self, out, aligner, ali, db)
                shogun_coverage(self, out, aligner, ali, db)
                shogun_function(self, out, aligner, ali, db, norm)

    if self.outputs['cmds']:
        self.outputs['cmds'] = combine_cmds + ali_cmds + self.outputs['cmds']


def woltka_aligments(self) -> dict:
    """Get the alignment paths per sample.

    Parameters
    ----------
    self : Commands class instance
        .inputs : dict
            Input files
        .outputs : dict
            All outputs

    Returns
    -------
    alis : dict
        Alignments
    """
    alis = {}
    for sample, sam_inputs in self.inputs.items():
        for db_pairing, sam in sam_inputs.items():
            if db_pairing[0] == 'wol':
                alis[sample] = sam
                io_update(self, i_f=sam)
    return alis


def woltka_write_map(self, alis: dict) -> str:
    """Write the mapping file that servers as input to Woltka.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to the output folder
        .outputs: dict
            All outputs
    alis : dict
        Alignments

    Returns
    -------
    woltka_map : str
        Path to the output woltka samples file.
    """
    woltka_map = '%s/samples.map' % self.dir
    for idx, sam in enumerate(alis.keys()):
        echo = 'echo -e "%s\\t%s"' % (sam, alis[sam])
        if idx:
            self.outputs['cmds'].append('%s >> %s' % (echo, woltka_map))
        else:
            self.outputs['cmds'].append('%s > %s' % (echo, woltka_map))
    return woltka_map


def woltka_gotus(self, woltka_map: str) -> str:
    """Get the Woltka command for the gotu classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    woltka_map : str
        Path to the output woltka samples file.

    Returns
    -------
    genomes_out : str
        Path to the output genomes per sample table.
    """
    genomes_out = '%s/genomes.tsv' % self.dir
    if not isfile(genomes_out):
        cmd = 'woltka gotu'
        cmd += ' -i %s' % woltka_map
        cmd += ' -o %s' % genomes_out
        self.outputs['cmds'].append(cmd)
        self.outputs['outs'].append(genomes_out)
        io_update(self, o_f=genomes_out)
    return genomes_out


def woltka_tax_cmd(self, woltka_map: str, database: str) -> str:
    """Get the taxonomic classification outputs and prepare the
    Woltka commands for this classification.

    Parameters
    ----------
    self : Commands class instance
        .out_dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    woltka_map : str
        Path to the Woltka input file.
    database : str
        WOL database path

    Returns
    -------
    tax_outmap : str
        Path to the folder containing the taxonomic maps.
    """
    tax_out, tax_outmap = '%s/taxa' % self.dir, '%s/taxmap' % self.dir
    tax_outputs = ['phylum', 'genus', 'species', 'none']
    tax_todo = []
    for tdx, tax_output in enumerate(tax_outputs):
        cur_tax_output = '%s/%s.tsv' % (tax_out, tax_output)
        self.outputs['cmds'].append(cur_tax_output)
        if not isfile(cur_tax_output):
            tax_todo.append(tax_output)
            io_update(self, o_f=cur_tax_output)

    taxid = '%s/taxonomy/taxid.map' % database
    nodes = '%s/taxonomy/nodes.dmp' % database
    names = '%s/taxonomy/names.dmp' % database
    io_update(self, i_f=[taxid, nodes, names])
    if len(tax_todo):
        cur_cmd = '\n# taxonomic\n'
        cur_cmd += 'woltka classify'
        cur_cmd += ' -i %s' % woltka_map
        cur_cmd += ' --map %s' % taxid
        cur_cmd += ' --nodes %s' % nodes
        cur_cmd += ' --names %s' % names
        cur_cmd += ' --rank %s' % ','.join(tax_outputs)
        cur_cmd += ' --add-rank'
        cur_cmd += ' --add-lineage'
        cur_cmd += ' --name-as-id'
        cur_cmd += ' --to-tsv'
        cur_cmd += ' --outmap %s' % tax_outmap
        cur_cmd += ' -o %s' % tax_out
        self.outputs['cmds'].append(cur_cmd)
    return tax_outmap


def woltka_classif_go(self, woltka_map: str, tax_outmap: str, database: str):
    """Get the taxonomic classification outputs and prepare the Woltka
    commands for this classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    woltka_map : str
        Path to the Woltka input file.
    tax_outmap : str
        Path to the folder containing the taxonomic maps.
    database : str
        WOL database path
    """
    coords = '%s/proteins/coords.txt.xz' % database
    uniref_map = '%s/function/uniref/uniref.map.xz' % database
    uniref_names = '%s/function/uniref/uniref.names.xz' % database
    io_update(self, i_f=[coords, uniref_map, uniref_names])
    go_rt = '%s/function/go' % database
    gos = ['process', 'function', 'component']
    woltka_fun_out = '%s/go' % self.dir
    for go in gos:
        cmd = '\n# %s [no stratification]\n' % go
        cmd += 'woltka classify'
        cmd += ' -i %s' % woltka_map
        cmd += ' --coords %s' % coords
        cmd += ' --map-as-rank'
        cmd += ' --rank %s' % go
        cmd += ' --map %s' % uniref_map
        cmd += ' --to-tsv'
        cur_map = '%s/%s.map.xz' % (go_rt, go)
        cmd += ' --map %s' % cur_map
        io_update(self, i_f=cur_map)
        cur_out = '%s/%s.tsv' % (woltka_fun_out, go)
        cmd += ' -o %s' % cur_out
        if not isfile(cur_out):
            self.outputs['cmds'].append(cmd)
            io_update(self, o_f=cur_out)
        self.outputs['outs'].append(cur_out)

    stratifications = ['phylum', 'family', 'genus', 'species']
    for stratification in stratifications:
        woltka_fun_out = '%s/go_%s' % (self.dir, stratification)
        for go in gos:
            cmd = '\n# %s [%s]\n' % (go, stratification)
            cmd += 'woltka classify'
            cmd += ' -i %s' % woltka_map
            cmd += ' --coords %s' % coords
            cmd += ' --map-as-rank'
            cmd += ' --rank %s' % go
            cmd += ' --stratify %s/%s' % (tax_outmap, stratification)
            cmd += ' --map %s' % uniref_map
            cmd += ' --to-tsv'
            cur_map = '%s/%s.map.xz' % (go_rt, go)
            cmd += ' --map %s' % cur_map
            go_out = '%s/%s.tsv' % (woltka_fun_out, go)
            cmd += ' -o %s' % go_out
            if not isfile(go_out):
                self.outputs['cmds'].append(cmd)
                io_update(self, o_f=go_out)
            self.outputs['outs'].append(go_out)


def woltka_classif_genes(self, woltka_map: str, database: str):
    """Get the Woltka commands for the gene-level classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    woltka_map : str
        Path to the Woltka input file.
    database : str
        WOL database path

    Returns
    -------
    genes : str
        Path to the genes classification output.
    """
    coords = '%s/proteins/coords.txt.xz' % database
    genes = '%s/wol_genes.biom' % self.dir
    if not isfile(genes):
        cmd = '\n# per gene\n'
        cmd += 'woltka classify'
        cmd += ' -i %s' % woltka_map
        cmd += ' --coords %s' % coords
        cmd += ' -o %s' % genes
        self.outputs['cmds'].append(cmd)
        io_update(self, o_f=genes)
    else:
        io_update(self, i_f=genes)
    self.outputs['outs'].append(genes)
    return genes


def woltka_classif_uniref(self, genes: str, database: str):
    """Get the Woltka commands for the uniref-level classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    genes : str
        Path to the genes classification output.
    database : str
        WOL database path

    Returns
    -------
    uniref : str
        Path to the uniref classification.
    """
    uniref_map = '%s/function/uniref/uniref.map.xz' % database
    uniref_names = '%s/function/uniref/uniref.names.xz' % database
    uniref = '%s/wol_uniref.biom' % self.dir
    if not isfile(uniref):
        cmd = 'woltka tools collapse'
        cmd += ' --input %s' % genes
        cmd += ' --map %s' % uniref_map
        cmd += ' --names %s' % uniref_names
        cmd += ' --output %s' % uniref
        self.outputs['cmds'].append(cmd)
        io_update(self, o_f=uniref)
    else:
        io_update(self, i_f=uniref)
    self.outputs['outs'].append(uniref)
    return uniref


def woltka_classif_eggnog(self, uniref: str, database: str):
    """Get the Woltka commands for the eggnog-level classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    uniref : str
        Path to the uniref classification.
    database : str
        WOL database path
    """
    biom = '%s/eggnog/wol_eggnog.biom' % self.dir
    if not isfile(biom):
        cmd = 'woltka tools collapse'
        cmd += '--input %s' % uniref
        cmd += ' --map %s/function/eggnog/eggnog.map.xz' % database
        cmd += ' --output %s\n\n' % biom
        self.outputs['cmds'].append(cmd)
        io_update(self, o_f=biom)
    else:
        io_update(self, i_f=biom)
    tsv = '%s.tsv' % splitext(biom)[0]
    if not isfile(tsv):
        cmd = 'biom convert -i %s -o %s.tmp --to-tsv\n' % (biom, tsv)
        cmd += 'tail -n +2 %s.tmp > %s\n' % (tsv, tsv)
        cmd += 'rm %s.tmp\n' % tsv
        self.outputs['cmds'].append(cmd)
        io_update(self, o_f=tsv)
    self.outputs['outs'].extend([tsv, biom])


def woltka_classif_cazy(self, genes: str, database: str):
    """Get the Woltka commands for the cazy-level classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .databases
            All databases class instance
        .outputs: dict
            All outputs
    genes : str
        Path to the genes classification.
    database : str
        WOL database path
    """
    cazy_map = '%s/function/cazy/3tools.txt' % database
    biom = '%s/cazy/wol_cazy.biom' % self.dir
    if not isfile(biom):
        cmd = 'woltka tools collapse'
        cmd += '--input %s' % genes
        cmd += ' --map %s' % cazy_map
        cmd += ' --output %s\n\n' % biom
        self.outputs['cmds'].append(cmd)
        io_update(self, o_f=biom)
    else:
        io_update(self, i_f=biom)
    tsv = '%s.tsv' % splitext(biom)[0]
    if not isfile(tsv):
        cmd = 'biom convert -i %s -o %s.tmp --to-tsv\n' % (biom, tsv)
        cmd += 'tail -n +2 %s.tmp > %s\n' % (tsv, tsv)
        cmd += 'rm %s.tmp\n' % tsv
        self.outputs['cmds'].append(cmd)
        io_update(self, o_f=tsv)
    self.outputs['outs'].extend([biom, tsv])


def woltka_classif_metacyc(self, genes: str, database: str):
    """Get the Woltka commands for the metacyc-level classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    genes : str
        Path to the genes classification.
    database : str
        WOL database path
    """
    metacyc_dir = '%s/function/metacyc' % database
    metacyc = [('protein', 'protein_name.txt', 'protein.map.xz'),
               ('enzrxn', 'enzrxn_name.txt', 'protein-to-enzrxn.txt'),
               ('reaction', 'reaction_name.txt', 'enzrxn-to-reaction.txt'),
               ('pathway', 'pathway_name.txt', 'reaction-to-pathway.txt'),
               ('super_pathway', '', 'pathway-to-super_pathway.txt'),
               ('regulation', '', 'enzrxn-to-regulation.txt'),
               ('regulator', '', 'regulation-to-regulator.txt'),
               ('ec', '', 'reaction-to-ec.txt')]
    files = [genes]
    woltka_fun_out = '%s/metacyc' % self.dir
    cmd = ''
    for idx, (level, names, maps) in enumerate(metacyc):
        input_fp = files[idx]
        if not input_fp.endswith('.biom'):
            input_biom = '%s.biom' % splitext(input_fp)[0]
            if not isfile(input_biom):
                cmd += 'biom convert'
                cmd += ' -i %s' % input_fp
                cmd += ' -o %s' % input_biom
                cmd += ' --to-hdf5'
                cmd += ' --table-type="OTU table"\n'
                io_update(self, o_f=input_biom)
            else:
                io_update(self, i_f=input_biom)
        else:
            input_biom = input_fp
        cmd += '\n# %s [no stratification]\n' % level
        biom = '%s/metacyc-%s_pre.biom' % (woltka_fun_out, level)
        if not isfile(biom):
            cmd += 'woltka tools collapse'
            if names:
                cmd += ' --input %s' % input_biom
                cmd += ' --names %s/%s' % (metacyc_dir, names)
            else:
                if level == 'super_pathway':
                    cmd += ' --input %s' % files[4]
                elif level == 'regulation':
                    cmd += ' --input %s' % files[2]
                elif level == 'regulator':
                    cmd += ' --input %s' % files[6]
                elif level == 'ec':
                    cmd += ' --input %s' % files[3]
                else:
                    cmd += ' --input %s' % files[4]
            cmd += ' --map %s/%s' % (metacyc_dir, maps)
            cmd += ' --output %s\n' % biom
            io_update(self, o_f=biom)
        else:
            io_update(self, i_f=biom)
        tsv = '%s/metacyc-%s_pre.tsv' % (woltka_fun_out, level)
        if not isfile(tsv):
            cmd += 'biom convert'
            cmd += ' -i %s' % biom
            cmd += ' -o %s.tmp' % tsv
            cmd += ' --to-tsv\n'
            cmd += 'tail -n +2 %s.tmp > %s\n' % (tsv, tsv)
            cmd += 'rm %s.tmp\n' % tsv
            io_update(self, o_f=tsv)
        else:
            io_update(self, i_f=tsv)
        files.extend([biom, tsv])
    if cmd:
        self.outputs['cmds'].append(cmd)
    self.outputs['outs'].extend(files)


def woltka_classif_kegg(self, uniref: str, database: str):
    """Get the Woltka commands for the kegg-level classification.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SHOGUN.
        .outputs: dict
            All outputs
    uniref : str
        Path to the uniref classification.
    database : str
        WOL database path
    """
    ko_names_maps = [
        ('ko', 'ko.name', 'ko.map.xz', ''),
        ('query', '', '', ''),
        ('ko-cog', '', 'ko-to-cog.txt', 'ko'),
        ('ko-disease', 'disease_name.txt', 'ko-to-disease.txt', 'ko'),
        ('ko-ec', '', 'ko-to-ec.txt', 'ko'),
        ('ko-go', '', 'ko-to-go.txt', 'ko'),
        ('ko-module', 'module_name.txt', 'ko-to-module.txt', 'ko'),
        ('ko-pathway', 'pathway_name.txt', 'ko-to-pathway.txt', 'ko'),
        ('ko-reaction', 'reaction_name.txt', 'ko-to-reaction.txt', 'ko'),
        ('ko-module-class', '', 'module-to-class.txt', 'ko-module'),
        ('ko-module-compound', 'compound_name.txt',
         'module-to-compound.txt', 'ko-module'),
        ('ko-module-ko', 'ko_name.txt', 'module-to-ko.txt', 'ko-module'),
        ('ko-module-pathway', 'pathway_name.txt',
         'module-to-pathway.txt', 'ko-module'),
        ('ko-module-reaction', 'reaction_name.txt',
         'module-to-reaction.txt', 'ko-module'),
        ('ko-pathway-class', '', 'pathway-to-class.txt', 'ko-pathway'),
        ('ko-pathway-compound', 'compound_name.txt',
         'pathway-to-compound.txt', 'ko-pathway'),
        ('ko-pathway-disease', 'disease_name.txt',
         'pathway-to-disease.txt', 'ko-pathway'),
        ('ko-pathway-ko', 'ko_name.txt', 'pathway-to-ko.txt', 'ko-pathway'),
        ('ko-pathway-module', 'module_name.txt',
         'pathway-to-module.txt', 'ko-pathway'),
        ('ko-reaction-compound', 'compound_name.txt',
         'reaction-to-compound.txt', 'ko-reaction'),
        ('ko-reaction-enzyme', '', 'reaction-to-enzyme.txt', 'ko-reaction'),
        ('ko-reaction-ko', 'ko_name.txt', 'reaction-to-ko.txt', 'ko-reaction'),
        ('ko-reaction-left_compound', 'compound_name.txt',
         'reaction-to-left_compound.txt', 'ko-reaction'),
        ('ko-reaction-module', 'module_name.txt',
         'reaction-to-module.txt', 'ko-reaction'),
        ('ko-reaction-pathway', 'pathway_name.txt',
         'reaction-to-pathway.txt', 'ko-reaction'),
        ('ko-reaction-rclass', 'rclass_name.txt',
         'reaction-to-rclass.txt', 'ko-reaction'),
        ('ko-reaction-right_compound', 'compound_name.txt',
         'reaction-to-right_compound.txt', 'ko-reaction')]
    cmd = ''
    files = []
    kegg_maps = '%s/kegg_queried' % self.dir
    for (level, name, maps, prev) in ko_names_maps:
        if maps:
            biom = '%s/kegg/kegg-%s_pre.biom' % (self.dir, level)
            tsv = '%s/kegg/kegg-%s_pre.tsv' % (self.dir, level)
            if not prev:
                if not isfile(tsv):
                    cmd += 'woltka tools collapse'
                    cmd += ' --input %s' % uniref
                    cmd += ' --names %s/function/kegg/%s' % (database, name)
                    cmd += ' --map %s/function/kegg/%s' % (database, maps)
                    cmd += ' --output %s\n' % biom

                    cmd += ' biom convert -i %s' % biom
                    cmd += ' -o %s.tmp --to-tsv\n' % tsv
                    cmd += ' tail -n +2 %s.tmp\n' % tsv
                    cmd += ' > %s\n' % tsv
                    cmd += ' rm %s.tmp\n' % tsv
                    io_update(self, o_f=[biom, tsv])
                else:
                    io_update(self, i_f=biom)
            else:
                input_fp = '%s/kegg/kegg-%s.biom' % (self.dir, level)
                if not isfile(tsv):
                    cmd += 'woltka tools collapse'
                    cmd += ' --input %s' % input_fp
                    if name:
                        cmd += ' --names %s/%s' % (kegg_maps, name)
                    cmd += ' --map %s/%s' % (kegg_maps, maps)
                    cmd += ' --output %s\n' % biom
                    cmd += 'biom convert -i %s\n' % biom
                    cmd += ' -o %s.tmp --to-tsv\n' % tsv
                    cmd += 'tail -n +2 %s.tmp > %s\n' % (tsv, tsv)
                    cmd += 'rm %s.tmp\n\n' % tsv
                    io_update(self, o_f=[biom, tsv])
                else:
                    io_update(self, i_f=biom)
        else:
            if not isfile('%s/kegg_info.txt' % kegg_maps):
                cmd += 'cd %s\n' % kegg_maps
                cmd += 'cp %s %s/%s\n' % (tsv, kegg_maps, basename(tsv))
                kegg_query = '%s/wol/kegg_query.py' % RESOURCES
                cmd += 'python3 %s %s\n' % (kegg_query, basename(tsv))
                io_update(self, o_d=kegg_maps)
            else:
                io_update(self, i_d=kegg_maps)
    if cmd:
        self.outputs['cmds'].append(cmd)
    self.outputs['outs'].extend(files)


def woltka(self) -> None:
    """Prepare the command and collect outputs, io and dirs for Woltka.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .databases
            All databases class instance
    """
    database = self.databases.paths['wol']
    alis = woltka_aligments(self)
    woltka_map = woltka_write_map(self, alis)
    woltka_gotus(self, woltka_map)
    tax_outmap = woltka_tax_cmd(self, woltka_map, database)
    woltka_classif_go(self, woltka_map, tax_outmap, database)
    genes = woltka_classif_genes(self, woltka_map, database)
    uniref = woltka_classif_uniref(self, genes, database)
    woltka_classif_eggnog(self, uniref, database)
    woltka_classif_cazy(self, genes, database)
    woltka_classif_metacyc(self, genes, database)
    woltka_classif_kegg(self, uniref, database)


def get_midas_cmd(
        self,
        focus_dir: str,
        db: str,
        analysis: str,
        select: set = None) -> str:
    """Build the command line for MIDAS analysis.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .soft.params
            Parameters
    focus_dir : str
        Path to the output folder.
    db : str
        Path to MIDAS database.
    analysis : str
        MIDAS analysis (any of "species", "genes" or "snps").
    select : set
        Species names for which there is a reference in the database.

    Returns
    -------
    cmd : str
        Midas command line for the species level.
    """
    cmd = 'run_midas.py %s' % analysis
    cmd += ' %s' % focus_dir
    cmd += ' -1 %s' % self.inputs[self.sam][0]
    if len(self.inputs[self.sam]) > 1:
        cmd += ' -2 %s' % self.inputs[self.sam][1]
    cmd += ' -d %s' % db
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' --remove_temp'
    if analysis != 'species':
        cmd += ' --species_cov 1'
    if select:
        cmd += ' --species_id %s' % ','.join(list(select))
    return cmd


def midas_species(self, focus_dir, db, tax) -> None:
    if db == self.databases.paths['midas']:
        species_out = '%s/species' % focus_dir
        species_profile = '%s/species_profile.txt' % species_out
        if not self.config.force and isfile(species_profile):
            io_update(self, i_d=species_out)
        else:
            self.outputs['cmds'].append(get_midas_cmd(self, focus_dir, db, tax))
            io_update(self, i_f=self.inputs[self.sam], o_d=species_out)
        self.outputs['outs'].append(species_out)
        self.outputs['dirs'].append(species_out)


def midas_genus(self, focus_dir, genes_out, db, tax, select):
    if self.config.force or not isfile('%s/readme.txt' % genes_out):
        self.outputs['cmds'].append(
            get_midas_cmd(self, focus_dir, db, tax, select))
        io_update(self, o_d=genes_out)
    self.outputs['outs'].append(genes_out)
    self.outputs['dirs'].append(genes_out)


def midas_snps(self, focus_dir, snps_out, db, tax, select):
    if self.config.force or not isfile('%s/readme.txt' % snps_out):
        self.outputs['cmds'].append(
            get_midas_cmd(self, focus_dir, db, tax, select))
        io_update(self, o_d=snps_out)
    self.outputs['outs'].append(snps_out)
    self.outputs['dirs'].append(snps_out)


def get_species_select(self, db: str, species_list: str) -> set:
    """Get the species names for which there is a reference in the database.

    Parameters
    ----------
    self : Commands class instance
        .config
            Configurations
    db : str
        Database name.
    species_list : str
        Path to file containing list of species to focus on.

    Returns
    -------
    select : set
        Species names for which there is a reference in the database.
    """
    select = set()
    if species_list:
        if self.config.dev and not isfile(species_list):
            select.add('Escherichia')
            return select
        else:
            species = [x.strip() for x in open(species_list).readlines()]
        with open('%s/species_info.txt' % db) as f:
            for line in f:
                for genus_species in species:
                    if genus_species.replace(' ', '_') in line:
                        select.add(line.split('\t')[0])
    return select


def midas(self) -> None:
    """Create command lines for MIDAS for the current database's species focus.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .databases
            All databases class instance
        .config
            Configurations
    """
    for focus, (db, species_list) in self.soft.params['focus'].items():

        focus_dir = '%s/%s/%s' % (self.dir, focus, self.sam)
        midas_species(self, focus_dir, db, 'species')

        select = set(get_species_select(self, db, species_list))

        genes_out = '%s/genes' % focus_dir
        midas_genus(self, focus_dir, genes_out, db, 'genus', select)

        snps_out = '%s/snps' % focus_dir
        midas_snps(self, focus_dir, snps_out, db, 'snps', select)


def get_kraken2_db(self, db):
    if db == 'default':
        db_path = self.databases.paths['kraken2']
    elif db in self.databases.paths:
        db_path = '%s/kraken2' % self.databases.paths[db]
        if not self.config.dev and not isdir(db_path):
            sys.exit('[kraken2] Not a database: %s' % db_path)
    else:
        sys.exit('[kraken2] Database not found: %s' % db)
    return db_path


def get_kraken2_cmd(self, out: str, db_path: str, report: str, result: str):
    cmd = 'kraken2 '
    cmd += ' -db %s' % db_path
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --report %s/report.tsv' % report
    cmd += ' --confidence %s' % self.soft.params['confidence']
    if len(self.inputs[self.sam]) > 1:
        unclass = ['%s/unclassified_%s.fastq' % (out, r) for r in [1, 2]]
        classif = ['%s/classified_%s.fastq' % (out, r) for r in [1, 2]]
        cmd += ' --unclassified-out %s/unclassified#.fastq' % out
        cmd += ' --classified-out %s/classified#.fastq' % out
        cmd += ' --paired'
    else:
        unclass = ['%s/unclassified.fastq' % out]
        classif = ['%s/classified.fastq' % out]
        cmd += ' --unclassified-out %s/unclassified.fastq' % out
        cmd += ' --classified-out %s/classified.fastq' % out
    if self.inputs[self.sam][0].endswith('.gz'):
        cmd += ' --gzip-compressed'
        io_update(self, o_f=([
            '%s.gz' % x for x in unclass] + ['%s.gz' % x for x in classif]))
    else:
        io_update(self, o_f=(unclass + classif))
    cmd += ' %s > %s' % (' '.join(self.inputs[self.sam]), result)
    return cmd


def kraken2(self) -> None:
    """Create command lines for kraken2.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .databases
            All databases class instance
        .config
            Configurations
    """
    for db in self.soft.params['databases']:
        out = '%s/%s/%s' % (self.dir, self.sam, db)
        self.outputs['dirs'].append(out)
        report = '%s/report.tsv' % out
        result = '%s/result.tsv' % out
        if self.config.force or not isfile(result):
            db_path = get_kraken2_db(self, db)
            cmd = get_kraken2_cmd(self, out, db_path, report, result)
            self.outputs['outs'].append(result)
            self.outputs['cmds'].append(cmd)
            io_update(self, i_f=self.inputs[self.sam], o_f=[report, result])


def metaxa2(self) -> None:
    if self.soft.prev == 'kneaddata':
        fas = basename(self.inputs[self.sam][0]).split('_1.fastq')[0]
    else:
        fas = basename(self.inputs[self.sam][0]).split('_R1.fastq')[0]

    databases = {'greengenes': '/home/flejzerowicz/databases/metaxa2/gg',
                 'wol_ssu': '/home/flejzerowicz/databases/metaxa2/wl',
                 'wol_ssu_g': '/home/flejzerowicz/databases/metaxa2/wl_g'}
    io_update(self, i_f=self.inputs[self.sam])
    for db, db_path in databases.items():
        dir_db = self.dir + '/' + db
        if not isdir(dir_db):
            os.makedirs(dir_db)
        rad = dir_db + '/' + fas
        summary = '%s.summary.txt' % rad
        taxonomy = '%s.taxonomy.txt' % rad
        reltax = '%s.reltax.txt' % rad
        cmd = ''
        if self.config.force or not isfile(summary):
            cmd += 'metaxa2'
            for idx, fastq in enumerate(self.inputs[self.sam]):
                cmd += ' -%s %s' % ((idx + 1), fastq)
            cmd += ' -o %s' % rad
            if db_path:
                cmd += ' -p %s/HMMs' % db_path
                cmd += ' -d %s/blast' % db_path
            if idx:
                if fastq.endswith('fasta'):
                    cmd += ' -f paired-fasta'
                else:
                    cmd += ' -f paired-end'
            else:
                if fastq.endswith('fasta'):
                    cmd += ' -f fasta'
                else:
                    cmd += ' -f fastq'
            if fastq.endswith('gz'):
                cmd += ' -z gzip'
            cmd += ' --megablast %s' % self.soft.params['megablast']
            cmd += ' --align %s' % self.soft.params['align']
            cmd += ' --mode %s' % self.soft.params['mode']
            cmd += ' --plus %s' % self.soft.params['plus']
            cmd += ' --cpu %s' % self.soft.params['cpus']
            if self.soft.params['graphical']:
                cmd += ' --graphical'
            if self.soft.params['reltax']:
                cmd += ' --reltax'
            if db == 'wol_ssu_g':
                cmd += ' --taxlevel 8'
            else:
                cmd += ' --taxlevel 7'
            cmd += '\n'
            io_update(self, o_f=[summary, taxonomy, reltax])
        else:
            io_update(self, i_f=taxonomy)
        self.outputs['outs'].extend([summary, taxonomy, reltax])

        redist_rad = '%s.redist' % rad
        redist_taxonomy = '%s.taxonomy.summary.txt' % redist_rad
        if self.config.force or not isfile(redist_taxonomy):
            cmd += '\nmetaxa2_ttt'
            cmd += ' -i %s' % taxonomy
            cmd += ' -o %s' % redist_rad
            cmd += ' -r 0.8'
            cmd += ' -d 0.7'
            io_update(self, o_f=redist_taxonomy)

        if cmd:
            self.outputs['cmds'].append(cmd)
        self.outputs['outs'].append(redist_taxonomy)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import unittest
from unittest.mock import patch

import os
import subprocess

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

import pkg_resources
from os.path import isdir, isfile, splitext
from metagenomix._io_utils import (
    read_yaml, get_fastq_header, get_cat_zcat, edit_fastq_cmd, mkdr, get_chunks,
    get_pfam_file, get_hmm_dat, get_pfams_cmd, reads_lines)

RESOURCES = pkg_resources.resource_filename('metagenomix', 'resources')
FOLDER = pkg_resources.resource_filename('metagenomix', 'tests/unittests')


class TestIOUtils(unittest.TestCase):

    def setUp(self) -> None:

        self.fastq_fp = '%s/fastqs/file.fastq' % FOLDER
        self.fastq_gz_fp = '%s.fastq.gz' % splitext(self.fastq_fp)[0]
        o = open(self.fastq_fp, 'w')
        o.write("@A.1.1 1 length=10\n")
        o.write("CTCCTGTGCC\n")
        o.write("+A.1.1 1 length=10\n")
        o.write("GGGGGGIIII\n")
        o.write("@A.2.1 2 length=10\n")
        o.write("GTCAGAGATA\n")
        o.write("+A.2.1 2 length=10\n")
        o.write("AGAAAGGGGG\n")
        o.close()
        subprocess.call(('gzip --keep %s' % self.fastq_fp).split())

        self.to_format = '%s/to_format.dat' % FOLDER
        self.formatted = '%s/formatted.tsv' % FOLDER
        o = open(self.to_format, 'w')
        o.write('# WHATEVER\n')
        o.write('#=GF AA  X\n')
        o.write('#=GF BB  X\n')
        o.write('//\n')
        o.write('# SOEVER\n')
        o.write('#=GF BB  Y\n')
        o.write('#=GF CC  Y\n')
        o.write('//\n')
        o.close()
        subprocess.call(('gzip --keep %s' % self.to_format).split())

        self.empty_fp = '%s/empty.tsv' % FOLDER
        o = open(self.empty_fp, 'w')
        o.close()

    def test_read_yaml(self):
        self.assertEqual({}, read_yaml(None))
        self.assertEqual({}, read_yaml(''))

        res = {'default': {'time': "48", 'nodes': "1", 'cpus': "1",
                           'mem_num': "5", 'mem_dim': "gb", 'env': "mg"}}
        file_path = '%s/run_params.yml' % RESOURCES
        self.assertEqual(res, read_yaml(file_path))

        file_path = 'not_a_file.yml'
        with self.assertRaises(IOError) as exc:
            read_yaml(file_path)
        self.assertEqual('No yaml file "%s"' % file_path, str(exc.exception))

    def test_get_fastq_header(self):
        first_line = get_fastq_header(self.fastq_fp)
        self.assertEqual(first_line, "@A.1.1 1 length=10")
        first_line = get_fastq_header(self.fastq_gz_fp)
        self.assertEqual(first_line, "@A.1.1 1 length=10")

    def test_get_cat_zcat(self):
        cat = get_cat_zcat(self.fastq_fp)
        self.assertEqual(cat, 'cat')
        cat = get_cat_zcat(self.fastq_gz_fp)
        self.assertEqual(cat, 'gzcat')

    def test_edit_fastq_cmd(self):
        ref_cmd = 'cat %s | ' % self.fastq_fp
        ref_cmd += "awk '{ if (NR%4==1) { print $1\"/1\" } "
        ref_cmd += "else if (NR%2 == 1) { print \"+\" } "
        ref_cmd += "else { print } }' | gzip > %s_renamed\n" % self.fastq_fp
        ref_cmd += "mv %s_renamed %s.gz\n" % (self.fastq_fp, self.fastq_fp)
        cmd = edit_fastq_cmd(self.fastq_fp, 1, 'illumina')
        self.assertEqual(ref_cmd, cmd)

        ref_cmd = ref_cmd.replace("/1", "/2")
        cmd = edit_fastq_cmd(self.fastq_fp, 2, 'illumina')
        self.assertEqual(ref_cmd, cmd)

        ref_cmd = 'gzcat %s | ' % self.fastq_gz_fp
        ref_cmd += "awk '{ if (NR%4==1) { print $1\"/1\" } "
        ref_cmd += "else if (NR%2 == 1) { print \"+\" } "
        ref_cmd += "else { print } }' | gzip > %s_renamed\n" % self.fastq_gz_fp
        ref_cmd += "mv %s_renamed %s\n" % (self.fastq_gz_fp, self.fastq_gz_fp)
        cmd = edit_fastq_cmd(self.fastq_gz_fp, 1, 'illumina')
        self.assertEqual(ref_cmd, cmd)

        ref_cmd = ref_cmd.replace("/1", "/2")
        cmd = edit_fastq_cmd(self.fastq_gz_fp, 2, 'illumina')
        self.assertEqual(ref_cmd, cmd)

        ref_cmd = 'cat %s | ' % self.fastq_fp
        ref_cmd += "awk '{ if (NR%4==1) { gsub(\".1$\",\"/1\",$1) "
        ref_cmd += "else if (NR%2 == 1) { print \"+\" } "
        ref_cmd += "else { print } }' | gzip > %s_renamed\n" % self.fastq_fp
        ref_cmd += "mv %s_renamed %s.gz\n" % (self.fastq_fp, self.fastq_fp)
        cmd = edit_fastq_cmd(self.fastq_fp, 1, 'ebi')
        self.assertEqual(ref_cmd, cmd)

        ref_cmd = ref_cmd.replace(".1$", ".2$").replace("/1", "/2")
        cmd = edit_fastq_cmd(self.fastq_fp, 2, 'ebi')
        self.assertEqual(ref_cmd, cmd)

        ref_cmd = 'gzcat %s | ' % self.fastq_gz_fp
        ref_cmd += "awk '{ if (NR%4==1) { gsub(\".1$\",\"/1\",$1) "
        ref_cmd += "else if (NR%2 == 1) { print \"+\" } "
        ref_cmd += "else { print } }' | gzip > %s_renamed\n" % self.fastq_gz_fp
        ref_cmd += "mv %s_renamed %s\n" % (self.fastq_gz_fp, self.fastq_gz_fp)
        cmd = edit_fastq_cmd(self.fastq_gz_fp, 1, 'ebi')
        self.assertEqual(ref_cmd, cmd)

        ref_cmd = ref_cmd.replace(".1$", ".2$").replace("/1", "/2")
        cmd = edit_fastq_cmd(self.fastq_gz_fp, 2, 'ebi')
        self.assertEqual(ref_cmd, cmd)

    def test_mkdr(self):
        folder = '%s/test_folder' % FOLDER
        self.assertFalse(isdir(folder))
        mkdr(folder)
        self.assertTrue(isdir(folder))
        os.rmdir(folder)
        self.assertFalse(isdir(folder))

    def test_get_chunks(self):
        ref_chunks = [[1, 2], [3, 4], [5]]
        chunks = get_chunks([1, 2, 3, 4, 5], 2)
        self.assertEqual(chunks, ref_chunks)

        ref_chunks = [[1, 2, 3], [4, 5]]
        chunks = get_chunks([1, 2, 3, 4, 5], 0, 3)
        self.assertEqual(chunks, ref_chunks)

        ref_chunks = [[1, 2, 3, 4], [5]]
        chunks = get_chunks([1, 2, 3, 4, 5], 0, 4)
        self.assertEqual(chunks, ref_chunks)

        ref_chunks = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        chunks = get_chunks([1, 2, 3, 4, 5, 6, 7, 8, 9], 3, 4)
        self.assertEqual(chunks, ref_chunks)

        ref_chunks = [[1, 2, 3, 4], [5, 6, 7, 8], [9]]
        chunks = get_chunks([1, 2, 3, 4, 5, 6, 7, 8, 9], 0, 4)
        self.assertEqual(chunks, ref_chunks)

        ref_chunks = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        chunks = get_chunks([1, 2, 3, 4, 5, 6, 7, 8, 9], 0, 3)
        self.assertEqual(chunks, ref_chunks)

    @patch('builtins.input', return_value='no')
    def test_get_pfam_file(self, input):
        ret = get_pfam_file(self.fastq_fp)
        self.assertTrue(ret)

    def test_get_hmm_dat(self):
        ret = pd.DataFrame({'AA': ['X', np.nan],
                            'BB': ['X', 'Y'],
                            'CC': [np.nan, 'Y']})
        ret_pd = get_hmm_dat('%s.gz' % self.to_format, self.formatted)
        assert_frame_equal(ret, ret_pd)
        self.assertTrue(isfile(self.formatted))

        ret_pd = get_hmm_dat('%s.gz' % self.to_format, self.formatted)
        assert_frame_equal(ret, ret_pd)
        os.remove(self.formatted)

    def test_get_pfams_cmd(self):

        self.assertFalse(isdir('./n'))
        pfam_key_pd = pd.DataFrame()
        ret_d, ret_cmd = get_pfams_cmd('HMM', pfam_key_pd, '.', 'n')
        self.assertTrue(isdir('./n'))
        self.assertEqual({}, ret_d)
        self.assertEqual('', ret_cmd)

        pfam_key_pd = pd.DataFrame({'AC': ['A', 'B'],
                                    'DE': ['1', '2']})
        d = {'A__1': ['./n/A__1.hmm', './n/A__1.dmnd'],
             'B__2': ['./n/B__2.hmm', './n/B__2.dmnd']}
        hmm1 = 'hmmfetch HMM A >> ./n/A__1.hmm\n'
        dmnd1 = 'cp ./fastas/A.fa ./fastas/A__1.fa\n'
        dmnd1 += 'diamond makedb --in ./fastas/A__1.fa -d ./n/A__1.dmnd\n'
        hmm2 = 'hmmfetch HMM B >> ./n/B__2.hmm\n'
        dmnd2 = 'cp ./fastas/B.fa ./fastas/B__2.fa\n'
        dmnd2 += 'diamond makedb --in ./fastas/B__2.fa -d ./n/B__2.dmnd\n'
        ret_d, ret_cmd = get_pfams_cmd('HMM', pfam_key_pd, '.', 'n')
        self.assertEqual(d, ret_d)
        cmd = hmm1 + dmnd1 + hmm2 + dmnd2
        self.assertEqual(cmd, ret_cmd)

        arefiles = [('./n/A__1.hmm', hmm1), ('./n/B__2.hmm', hmm2),
                    ('./n/A__1.dmnd', dmnd1), ('./n/B__2.dmnd', dmnd2)]
        for (fp, c) in arefiles:
            o = open(fp, 'w')
            o.close()
            cmd = cmd.replace(c, '')
            ret_d, ret_cmd = get_pfams_cmd('HMM', pfam_key_pd, '.', 'n')
            self.assertEqual(cmd, ret_cmd)
        for (fp, _) in arefiles:
            os.remove(fp)
        os.rmdir('./n')

    def test_reads_lines(self):

        ret = reads_lines(self.empty_fp)
        self.assertEqual(ret, [])
        ret = reads_lines(self.to_format)
        self.assertEqual(ret, ['# WHATEVER', '#=GF AA  X', '#=GF BB  X', '//',
                               '# SOEVER', '#=GF BB  Y', '#=GF CC  Y', '//'])

    def tearDown(self) -> None:
        os.remove(self.fastq_fp)
        os.remove(self.fastq_gz_fp)
        os.remove(self.to_format)
        os.remove('%s.gz' % self.to_format)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import glob
import sys
from skbio.io import read

from os.path import basename, dirname, isdir, isfile, splitext
from metagenomix._io_utils import (
    get_out_dir, write_hmms, io_update, reads_lines)

from metagenomix._cmds import caller


def prodigal_cmd(self, contigs_fp: str, out_dir: str, group: str = '') -> tuple:
    """Create command lines for Prodigal.

    self : Commands class instance
        .dir : str
            Path to pipeline output folder for prodigal
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    contigs_fp : str
        Contigs to predict
    out_dir : str
        Output directory
    group
        Pooling group
    """
    gbk = '%s/gene.coords.gbk' % out_dir
    proteins = '%s/protein.translations.fasta' % out_dir
    genes = '%s/potential.starts.fasta' % out_dir
    cmd = 'prodigal'
    cmd += ' -i %s' % contigs_fp
    cmd += ' -o %s' % gbk
    cmd += ' -a %s' % proteins
    cmd += ' -s %s' % genes
    cmd += ' -p %s' % self.soft.params['procedure']
    outputs = [gbk, proteins, genes]
    return cmd, outputs


def prodigal(self):
    """Create command lines for Prodigal.

    self : Commands class instance
        .dir : str
            Path to pipeline output folder for prodigal
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.pool in self.pools:
        for group in self.pools[self.pool]:
            contigs = self.inputs[self.pool][group][1]
            out_dir = '%s/%s/%s' % (self.dir, self.pool, group)
            cmd, outputs = prodigal_cmd(self, contigs, out_dir, group)
            self.outputs['outs'][group] = outputs
            if self.config.force or not isfile(outputs[0]):
                self.outputs['cmds'].setdefault(group, []).append(cmd)
                io_update(self, i_f=contigs, o_f=outputs, key=group)
    else:
        contigs = self.inputs[self.sam]
        out_dir = '%s/%s' % (self.dir, self.sam)
        cmd, outputs = prodigal_cmd(self, contigs, out_dir)
        self.outputs['outs'].extend(outputs)
        self.outputs['cmds'].append(cmd)
        io_update(self, i_f=contigs, o_f=outputs)


def macsyfinder_cmd(self, fp, o_dir, folder, group=None) -> tuple:
    cmd = ''
    outs = []
    io_update(self, i_f=fp, i_d=folder, key=group)
    self.outputs['dirs'].append(o_dir)
    for model in self.soft.params['models']:
        out_dir = '%s/%s' % (o_dir, model)
        outs.append(out_dir)
        self.outputs['dirs'].append(out_dir)
        io_update(self, o_d=out_dir, key=group)
        res = '%s/macsyfinder.log' % out_dir
        if self.config.force or not isfile(res):
            cmd += 'macsyfinder'
            cmd += ' --db-type %s' % self.soft.params['db_type']
            cmd += ' --replicon-topology %s' % self.soft.params['topology']
            cmd += ' --e-value-search %s' % self.soft.params['evalue']
            cmd += ' --coverage-profile %s' % self.soft.params['coverage']
            cmd += ' --out-dir %s' % out_dir
            cmd += ' --res-search-suffix _hmm.tsv'
            cmd += ' --res-extract-suffix _out.tsv'
            cmd += ' --worker %s' % self.soft.params['cpus']
            cmd += ' --sequence-db %s' % fp
            cmd += ' --models-dir %s/data/models' % folder
            cmd += ' --models %s all' % model
            cmd += ' --verbosity\n'
    return cmd, outs


def macsyfinder(self) -> None:
    """Create command lines for MacSyFinder.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for macsyfinder
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    folder = self.databases.paths['macsyfinder']
    if self.pool in self.pools:
        self.outputs['outs'] = {}
        for group in self.pools[self.pool]:
            o_dir, fp = get_out_dir(self, self.pool, group)
            cmd, outs = macsyfinder_cmd(self, fp, o_dir, folder, group)
            self.outputs['outs'][group] = outs
            if cmd:
                self.outputs['cmds'].setdefault(group, []).append(cmd)
    else:
        o_dir, fp = get_out_dir(self, self.sam)
        fp_edit = '%s_edit.fasta' % fp.replace('.fasta', '')
        cmd_edit = 'header_space_replace.py -i %s -o %s --n\n' % (fp, fp_edit)
        cmd, outs = macsyfinder_cmd(self, fp_edit, o_dir, folder)
        self.outputs['outs'].extend(outs)
        if cmd:
            cmd = cmd_edit + cmd
            self.outputs['cmds'].append(cmd)


def prepare_ioncom_inputs(self, ion_com: str, fp: str) -> str:
    """Create the preformatting command lines before running IonCom.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for IonCom
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    ion_com : str
        Path to the IonCom standalone
    fp : str
        Path to the output file
    """
    cmd = 'prepare_ioncom_inputs.py'
    cmd += ' -d %s -p %s -n 1' % (ion_com, fp)
    cmd += '%s/run_IonCom.pl' % ion_com
    return cmd


def ioncom(self) -> None:
    """Create command lines for IonCom.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for IonCom
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    # -------------------------------------------------------
    # --- WILL NEED TO BE REVISED AFTER I-TASSER DOWNLOAD ---
    # -------------------------------------------------------
    i_tasser_libs = self.databases.ioncom['itasser']
    i_tasser = self.soft.params['itasser']
    ion_com = '%s/IonCom_standalone' % self.databases.paths['ioncom']

    if self.pool in self.pools:
        for group in self.pools[self.pool]:
            o_dir, fp = get_out_dir(self, self.pool, group)

            cmd = prepare_ioncom_inputs(self, ion_com, fp)
            self.outputs['cmds'].setdefault(group, []).append(cmd)

            output = '%s/output' % ion_com
            cmd = 'for i in %s/*rehtml\n' % output
            cmd += 'do\n'
            # cmd += '    mkdir -p "%s/$(basename $(dirname "$i"))\n' % (
            #     output, out_dir)
            cmd += '    mv %s/*rehtml %s/.\n' % (output, o_dir)
            cmd += 'done\n'
            cmd += 'for i in %s/*/*/prob_*.txt\n' % output
            cmd += 'do\n'
            cmd += '    mkdir -p "$i"\n'
            cmd += 'done'
            self.outputs['cmds'].setdefault(group, []).append(cmd)

            self.outputs['dir'].append(o_dir)
            self.outputs['outs'].setdefault(self.pool, []).append(o_dir)
            io_update(self, i_f=fp, i_d=[ion_com, i_tasser, i_tasser_libs],
                      o_d=o_dir, key=group)


def integron_finder(self):
    """Create command lines for integron_finder.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for integron_finder
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.pool in self.pools:
        for group in self.pools[self.pool]:
            o_dir, fp = get_out_dir(self, self.pool, group)
            fp_out = fp.replace('.fasta',
                                '_len%s.fasta' % self.soft.params['min_length'])
            hmms_fp = write_hmms(self)

            cmd = 'filter_on_length.py'
            cmd += ' -i %s' % fp
            cmd += ' -o %s' % fp_out
            cmd += ' -t %s\n' % self.soft.params['min_length']

            cmd += 'integron_finder'
            if self.soft.params['local_max']:
                cmd += ' --local-max'
            if self.soft.params['promoter_attI']:
                cmd += ' --promoter-attI'
            if self.soft.params['mute']:
                cmd += ' --mute'
            if self.soft.params['pdf']:
                cmd += ' --pdf'
            if self.soft.params['gbk']:
                cmd += ' --gbk'
            if self.soft.params['union_integrases']:
                cmd += ' --union-integrases'

            cmd += ' --verbose'
            cmd += ' --outdir %s' % o_dir
            cmd += ' --cpu %s' % self.soft.params['cpus']
            if self.databases.hmms_dias:
                cmd += ' --func-annot'
                cmd += ' --path-func-annot %s' % hmms_fp
                io_update(self, i_f=hmms_fp, key=group)
            if self.soft.params['prot_file']:
                cmd += ' --prot-file %s' % self.soft.params['prot_file']
            if self.soft.params['attc_model']:
                cmd += ' --attc-model %s' % self.soft.params['attc_model']
                cmd += ' --evalue-attc %s' % self.soft.params['evalue_attc']
                cmd += ' --max-attc-size %s' % self.soft.params['max_attc_size']
                cmd += ' --min-attc-size %s' % self.soft.params['min_attc_size']
            if self.soft.params['topology_file']:
                cmd += ' --topology-file %s' % self.soft.params['topology_file']
            else:
                cmd += ' --%s' % self.soft.params['topology']

            cmd += ' %s' % fp_out

            self.outputs['dirs'].append(o_dir)
            self.outputs['outs'].setdefault(self.pool, []).append(o_dir)
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=fp, o_d=o_dir, key=group)













def custom_cmd(self, input_file, out_dir, dia_hmm):
    outs = {}
    for target, gene_hmm_dia in self.databases.hmms_dias.items():
        if target == 'cazy':
            continue
        for gene, (hmm, dia) in gene_hmm_dia.items():
            sam_dir = '%s/%s' % (out_dir, gene)
            self.outputs['dirs'].append(sam_dir)
            out = '%s/%s.tsv' % (sam_dir, self.sam)
            if not isfile(out):
                if dia_hmm == 'hmm':
                    stdout = '%s_hmmer.out' % splitext(out)[0]
                    io_update(self, i_f=hmm, o_f=[out, stdout], key=target)
                    cmd = 'hmmsearch'
                    cmd += ' --cut_tc'
                    cmd += ' --tblout %s' % out
                    cmd += ' -o %s' % stdout
                    cmd += ' --cpu %s' % self.soft.params['cpus']
                    cmd += ' %s %s' % (hmm, input_file)
                else:
                    io_update(self, i_f=dia, o_f=out, key=target)
                    tmp = '%s/%s_tmp' % (sam_dir, self.sam)
                    cmd = 'mkdir -p %s\n' % tmp
                    cmd += 'diamond blastp'
                    cmd += ' -d %s' % dia
                    cmd += ' -q %s' % input_file
                    cmd += ' -o %s' % out
                    cmd += ' -k 1 --id 80'
                    cmd += ' -p %s' % self.soft.params['cpus']
                    cmd += ' -t %s' % tmp
                self.outputs['cmds'].setdefault(target, []).append(cmd)
            else:
                io_update(self, i_f=out, key=target)
            out2 = out.replace('.tsv', '_contigs.tsv')
            if not isfile(out2):
                io_update(self, o_f=out2, key=target)
                cmd = 'extract_custom_searched_contigs.py'
                cmd += ' -i %s' % out
                cmd += ' -o %s' % out2
                self.outputs['cmds'].setdefault(target, []).append(cmd)
            outs.setdefault(target, []).extend([out, out2])
    return outs


def prep_custom(self, dia_hmm):
    self.outputs['outs'] = {}
    if self.pool in self.pools:
        for group in self.pools[self.pool]:
            self.sam = group
            o_dir, fp = get_out_dir(self, self.pool, group)
            outs = self.custom_cmd(fp, o_dir, dia_hmm)
            io_update(self, i_f=fp, key=group)
            self.outputs['outs'][group] = outs
    else:
        o_dir, fp = get_out_dir(self, self.sam)
        outs = self.custom_cmd(fp, o_dir, dia_hmm)
        io_update(self, i_f=fp)
        self.outputs['outs'] = outs


def prep_diamond_custom(self):
    self.prep_custom('dia')


def prep_hmmer_custom(self):
    self.prep_custom('hmm')




















def diamond_annot_cmd(self, o_dir, fp, tmp_dir, group=None) -> tuple:
    cmd, outs = '', []
    io_update(self, i_f=fp, key=group)
    self.soft.dirs.add(o_dir)
    k = self.soft.params['k']
    for db, dmnds in self.soft.params['databases'].items():
        for dmnd in dmnds:
            base_db = basename(dmnd).rstrip('.dmnd')
            out = '%s/%s_k%s_%s.tsv' % (o_dir, self.sam, k, base_db)
            outs.append(out)
            cmd += 'mkdir -p %s\n' % tmp_dir
            cmd += 'diamond blastp'
            cmd += ' -d %s' % dmnd
            cmd += ' -q %s' % fp
            cmd += ' -o %s' % out
            cmd += ' -k %s' % k
            cmd += ' -p %s' % self.soft.params['cpus']
            cmd += ' --id %s' % self.soft.params['identity']
            cmd += ' -t %s\n' % tmp_dir
            io_update(self, o_f=out, key=group)

    return cmd, outs


def diamond(self):
    tmp_dir = '$TMPDIR/diamond_annot_%s' % self.sam
    if self.pool in self.pools:
        for group in self.pools[self.pool]:
            o_dir, fp = get_out_dir(self, self.pool, group)
            cmd, outs = diamond_annot_cmd(self, o_dir, fp, tmp_dir, group)
            self.outputs['outs'].setdefault(self.pool, []).extend(outs)
            if cmd:
                self.outputs['cmds'].setdefault(group, []).append(cmd)

    else:
        o_dir, fp = get_out_dir(self, self.sam)
        cmd, outs = diamond_annot_cmd(self, o_dir, fp, tmp_dir)
        self.outputs['outs'].extend(outs)
        if cmd:
            self.outputs['cmds'].append(cmd)


def search(self) -> None:
    """Create command lines for custom searches based on DIAMOND or HMMER.

    Parameters
    ----------
    self : Commands class instance
        Contains all the attributes needed for binning on the current sample
    """
    # This function splits the name of the software and calls as function (
    # from this module) the part of the software name that follows the first
    # underscore, e.g. software "search_diamond" would call `diamond()`
    caller(self, __name__)


def get_antismash_cmd(self, fa, base, out):
    cmd = '\nantismash'
    for boolean in ['rre', 'asf', 'cassis', 'pfam2go', 'tigrfam', 'fullhmmer',
                    'cc-mibig', 'cb_general', 'smcog_trees', 'clusterhmmer',
                    'cb_subclusters', 'cb_knownclusters']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    cmd += ' --tta-threshold %s' % self.soft.params['--tta-threshold']
    cmd += ' --genefinding-tool %s' % self.soft.params['genefinding_tool']
    cmd += ' --html-description after_%s' % self.soft.prev
    cmd += ' --taxon %s' % self.soft.params['taxon']
    cmd += ' --cpus %s' % self.soft.params['cpus']
    cmd += ' --output-basename %s' % base
    cmd += ' --html-title "%s: %s"' % base
    cmd += ' --output-dir %s' % out
    cmd += ' %s\n' % fa
    return cmd


def antismash(self):
    if self.soft.prev != 'drep':
        sys.exit('[%s] Not possible (yet) after %s (only drep)' % (
            self.soft.name, self.soft.prev))
    for pool in self.pools:
        for algo, folder in self.inputs[pool].items():
            out_dir = '%s/%s/%s' % (self.dir, pool, algo)
            io_update(self, o_d=out_dir, key=pool)
            for fa in glob.glob('%s/*.fa' % folder):
                io_update(self, i_f=fa, key=pool)
                base = fa.split('/')[-1].replace('.fa', '')
                out = '%s/%s' % (out_dir, base)
                self.outputs['dirs'].append(out)
                if self.config.force or not glob.glob('%s/*' % out):
                    cmd = get_antismash_cmd(self, fa, base, out)
                    self.outputs['cmds'].setdefault(
                        (pool, algo), []).append(cmd)


def get_prokka_cmd(self, contigs, out_dir, prefix, config, cols):
    cmd = 'prokka'
    cmd += ' --mincontiglen %s' % self.soft.params['mincontiglen']
    cmd += ' --cpus %s' % self.soft.params['cpus']
    for boolean in ['metagenome', 'notrna', 'norrna', 'annotation']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    cmd += ' --force'
    for col in cols:
        if col and config[col]:
            cmd += ' --%s %s' % (col, config[col])
    cmd += ' --prefix %s' % prefix
    cmd += ' --outdir %s' % out_dir
    cmd += ' %s' % contigs
    return cmd


def get_prokka_config(self, cols):
    configs = []
    cols += ['proteins']
    if 'taxa' in self.soft.params:
        with open(self.soft.params['taxa']) as f:
            for ldx, line in enumerate(f):
                ls = line.strip('\n').split('\t')
                if not ldx:
                    continue
                configs.append({cols[idx]: val for idx, val in enumerate(ls)})
    return configs


def prokka(self):
    if self.pool in self.pools:
        for group in self.pools[self.pool]:
            out = '%s/%s/%s' % (self.dir, self.pool, group)
            contigs = self.inputs[self.pool][group][1]
            if self.soft.params['config']:
                cols = ['genus', 'species', 'strain', 'plasmid']
                configs = get_prokka_config(self, cols)
            else:
                cols = ['']
                configs = {'': 'no_config'}

            cmd = ''
            for config in configs:
                prefix = '_'.join([config[x] for x in cols if config[x]])
                if config['proteins']:
                    prefix += '_%s' % splitext(basename(config['proteins']))[0]
                file_out = '%s/%s.out' % (out, prefix)
                if self.config.force or not isfile(file_out):
                    cmd += get_prokka_cmd(self, contigs, out,
                                          prefix, config, cols)

            self.outputs['dirs'].append(out)
            self.outputs['outs'].setdefault(group, []).append(out)
            if cmd:
                self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=configs, o_d=out, key=group)


def get_barrnap_cmd(self, out, fasta):
    cmd = 'barrnap'
    cmd += ' --kingdom %s' % self.soft.params['kingdom']
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --reject %s' % self.soft.params['reject']
    cmd += ' --lencutoff %s' % self.soft.params['lencutoff']
    cmd += ' --evalue %s' % self.soft.params['evalue']
    if self.soft.params['incseq']:
        cmd += ' --incseq'
    cmd += ' --outseq %s' % out
    cmd += ' %s' % fasta
    return cmd


def get_ccmap_cmd(self, out, fasta):
    out_dir = dirname(out)
    cmd = ''
    if self.config.force:
        cmd += 'if [ -d "%s" ]; then rm -rf %s; fi\n' % (out_dir, out_dir)
    cmd += 'ccfind %s %s' % (fasta, out_dir)
    cmd += ' --terminal-fragment-size %s' % self.soft.params[
        'terminal_fragment_size']
    cmd += ' --min-percent-identity %s' % self.soft.params[
        'min_percent_identity']
    cmd += ' --min-aligned-length %s' % self.soft.params[
        'min_aligned_length']
    cmd += ' --ncpus %s' % self.soft.params['cpus']
    if self.soft.params['preserve_tmpdir']:
        cmd += ' --preserve-tmpdir'
    return cmd


def generic_on_fasta(self, tool):
    func = 'get_%s_cmd' % self.soft.name
    if self.pool in self.pools:
        for group in self.pools[self.pool]:
            fastas_d = {}
            if self.soft.prev == 'spades':
                fastas_d = {'': [self.inputs[self.pool][group][1]]}
            elif self.soft.prev == 'drep':
                fastas_d = self.inputs[self.pool][group]
            for _, fastas in fastas_d.items():
                for fasta in fastas:
                    out = '%s/%s/%s/%s.fas' % (self.dir, self.pool, group,
                                               splitext(basename(fasta))[0])
                    self.outputs['outs'].setdefault(
                        (self.pool, group), []).append(out)
                    self.outputs['dirs'].append(dirname(out))
                    io_update(self, i_f=fasta, o_d=out, key=group)
                    if tool == 'barrnap':
                        condition = isfile(out)
                    elif tool == 'ccmap':
                        condition = isdir(dirname(out))
                    if self.config.force or not condition:
                        cmd = globals()[func](self, out, fasta)
                        self.outputs['cmds'].setdefault(group, []).append(cmd)
    else:
        for fasta in self.inputs[self.sam]:
            out = '%s/%s/%s.fas' % (self.dir, self.sam,
                                    splitext(basename(fasta))[0])
            self.outputs['dirs'].append(dirname(out))
            self.outputs['outs'].append(out)
            io_update(self, i_f=fasta, o_d=out)
            if tool == 'barrnap':
                condition = isfile(out)
            elif tool == 'ccmap':
                condition = isdir(dirname(out))
            if self.config.force or not condition:
                cmd = globals()[func](self, out, fasta)
                self.outputs['cmds'].append(cmd)


def barrnap(self):
    generic_on_fasta(self, self.soft.name)


def ccmap(self):
    generic_on_fasta(self, self.soft.name)


def write_dbcan_subset(self, taxa: list, folder: str, name: str) -> str:
    """Write the fasta file suset to the target features and
    the command to make it a diamond database.

    Parameters
    ----------
    taxa : list
        Taxa from the user file.
    folder : str
        dbCAN-Seq folder.
    name : str
        Current subset name.

    Returns
    -------
    cmd : str
        Command to make the diamond db from the subsets fasta file.
    """
    path = self.databases['dbcan']
    cmd = ""
    fas_fp = '%s/%s.fa' % (folder, name)
    dia_fp = '%s.dmnd' % splitext(fas_fp)[0]
    io_update(self, i_f=fas_fp, o_f=dia_fp)
    if not isfile(dia_fp):
        with open(fas_fp, 'w') as o:
            taxon_found = False
            for taxon in taxa:
                meta_taxon_pd = self.databases.dbcan_meta.loc[
                    self.databases.dbcan_meta.genome_name.str.contains(taxon),:]
                if not meta_taxon_pd.shape[0]:
                    continue
                gcf_dir = '%s/dbCAN-seq/CAZyme_seq_list' % path
                if not isdir(gcf_dir):
                    os.makedirs(gcf_dir)
                for gcf in set(meta_taxon_pd.index):
                    gcf_fas = '%s/%s.fasta' % (gcf_dir, gcf)
                    if isfile(gcf_fas):
                        for e in read(gcf_fas, 'fasta'):
                            o.write('>%s\n%s\n' % (e.metadata['id'], e))
                        taxon_found = True
        if taxon_found:
            cmd = "diamond makedb --in %s -d %s\n" % (fas_fp, dia_fp)
    self.outputs['cmds'].append(cmd)


def set_dbcan_taxa(self) -> None:
    for name, fp in self.soft.params['taxa'].items():
        taxa = list(reads_lines(fp))
        if not taxa:
            continue
        folder = '%s/subsets' % self.dir
        if not isdir(folder):
            os.makedirs(folder)
        self.databases.cazys[name] = folder
        write_dbcan_subset(self, taxa, folder, name)


def cazy(self):
    set_dbcan_taxa(self)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
from metagenomix._io_utils import mkdr


def get_drep_bins(prev: str, pools: dict, inputs: dict) -> dict:
    """Get the genomes in an iterable format accommodating
    both the output of `metawrap_refine()` and of `metawrap_reassemble()`.

    Parameters
    ----------
    prev : str
        Name of the software preceding drep.
    pools : dict
        Samples per group as dispatched according to the pooling design.
    inputs : dict
        Outputs from the software preceding drep.

    Returns
    -------
    bins : dict
        Bins to dereplicate.
    """
    bins = {}
    for pool in pools:
        bins[pool] = {}
        for group, sam_paths in inputs[pool].items():
            if prev == 'metawrap_refine':
                bins[pool].setdefault('', []).append((group, sam_paths[1]))
            elif prev == 'metawrap_reassemble':
                for sam, paths in sam_paths.items():
                    for path in paths:
                        if path.endswith('strict'):
                            bins[pool].setdefault('strict', []).append(path)
                        elif path.endswith('permissive'):
                            bins[pool].setdefault('permissive', []).append(path)
            else:
                raise ValueError('No genome dereplication after %s' % prev)
    return bins


def get_drep_inputs(drep_dir: str, sam_paths: list):
    """Write the file containing the inputs bins to drep
    and the list of paths to these bins.

    Parameters
    ----------
    drep_dir :
        Path to the drep output folder.
    sam_paths : list
        List containing for each (group, bins_folder).

    Returns
    -------
    drep_in : str
        File containing the paths corresponding to each bin.
    paths : list
        List of paths corresponding to each bin.
    """
    paths = []
    mkdr(drep_dir)
    drep_in = '%s/input_genomes.txt' % drep_dir
    with open(drep_in, 'w') as o:
        for sam_path in sam_paths:
            for path in glob.glob('%s/*fa' % sam_path):
                paths.append(path)
                o.write('%s\n' % path)
    return drep_in, paths



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

def post_processes(self):
    if self.name.startswith('map__spades') or self.name in [
        'count_reads', 'humann2', 'midas', 'cazy', 'prodigal',
        'count_reads_grep', 'diamond_custom', 'hmmer_custom'
    ]:
        out_merge_samples = merge_samples(soft, output_paths)
        self.outputs['%s_merge' % self.name] = out_merge_samples
    if self.name in ['cazy']:
        out_cazy = cazy_post_processing(soft, soft_prev, output_paths)
        output_paths['%s_data' % self.name] = out_cazy
        cazy_sequence_analysis(soft, soft_prev, output_paths)
    if self.name in ['midas']:
        out_postprocess_samples = postprocess(soft, output_paths)
    if self.name in ['spades']:
        outs = contigs_per_sample(soft, output_paths)
        output_paths['%s_per_sample' % self.name] = outs
        outs = longest_contigs_per_sample(soft, output_paths)
        output_paths['%s_longest' % self.name] = outs
        prep_quast(soft, output_paths)
    if self.name in [
            'metawrap_analysis_reassemble_bins', 'metawrap_ref', 'yamb']:
        rename_bins(soft, output_paths)
    if self.name in ['drep']:
        faindex_contig_to_bin_drep(soft, output_paths)
        outs = longest_contigs_per_drep_bin(soft, output_paths)
        output_paths['%s_longest' % self.name] = outs
        outs = contigs_per_drep_bin(soft, output_paths)
        output_paths['%s_bins_contigs' % self.name] = outs
        outs = genes_per_mag(soft, output_paths)
        output_paths['%s_genes_per_mag' % self.name] = outs
    if self.name in ['yamb']:
        outs = yamb_across(soft, output_paths)
        output_paths['%s_maps' % self.name] = outs
    if self.name in ['read_mapping']:
        collect_reads_per_contig(soft, soft_prev, output_paths)

    return all_sh, output_paths, mem_n_u, procs, cnd, pooling_groups, IOs




# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import glob
import yaml
import pkg_resources

import pandas as pd
from skbio.tree import TreeNode
from os.path import basename, isdir, isfile, splitext

from metagenomix._io_utils import (mkdr, get_pfam_wget_cmd, get_hmm_dat,
                                   get_pfams_cmd)

RESOURCES = pkg_resources.resource_filename('metagenomix', 'resources')


class ReferenceDatabases(object):

    def __init__(self, config) -> None:
        self.config = config
        self.commands = {}
        self.cmds = {}
        self.paths = {}
        self.builds = {}
        self.database = ''
        self.pfams = {'terms': {}}
        self.hmms_dias = {}
        self.hmms_pd = pd.DataFrame()
        self.dbcan_meta = pd.DataFrame()
        self.cazys = {}
        self.gtdb = {}
        self.wol = {}
        self.formats = [
            'bbmap', 'blastn', 'bowtie2', 'burst', 'centrifuge', 'diamond',
            'hmmer', 'kraken2', 'minimap2', 'qiime2', 'utree']
        self.valid_databases = set()

    def run(self) -> None:
        self.get_formats()
        self.validate_databases()
        self.set_databases()

    def get_formats(self):
        """Override init (default) database formats with those from yaml file"""
        if 'formats' in self.config.databases:
            self.formats = self.config.databases['formats']

    def validate_databases(self) -> None:
        """Show and validate the presence of databases.
        Whether there are valid databases in the user-defined paths.
        """
        if len(self.config.databases):
            x = '%s\n' % ('#' * (11 + len(self.config.databases_yml)))
            print('\n%sDatabases: %s\n%s' % (x, self.config.databases_yml, x))
            for db, path in sorted(self.config.databases.items()):
                if db == 'formats':  # "formats" is not a database
                    continue
                if not path or not isinstance(path, str):  # must have a "path"
                    print('  - %s: path (char. string) missing (ignored)' % db)
                    continue
                if not self.config.dev:
                    if not isdir(path):  # not-found database will be ignored
                        print("  - %s: can't find %s (ignored)" % (db, path))
                    else:  # print database input if it found to exist
                        self.valid_databases.add(db)
                        print('  + %s' % yaml.dump({db: path}), end='')
                else:  # print database input if it found to exist in dev mode
                    self.valid_databases.add(db)
                    print('  + %s' % yaml.dump({db: path}), end='')
        else:
            print('No database passed to option `-d`')

    def set_databases(self) -> None:
        for database in sorted(self.valid_databases):
            self.database = database
            if hasattr(self, "set_%s" % database):
                # this is true for the following databases:
                #  - wol
                #  - mar
                #  - pfam
                #  - dbcan
                #  - ioncom
                #  - shogun
                getattr(self, "set_%s" % database)()
            else:
                self.set_path()

    def set_path(self) -> None:
        self.paths[self.database] = self.config.databases[self.database]
        self.builds[self.database] = self.get_db_formats()

    def register_command(self) -> None:
        self.commands.setdefault(self.database, {}).update(dict(self.cmds))
        self.cmds = {}

    # Below are the database-specific functions, potentially to make builds

    def get_db_formats(self):
        path = self.paths[self.database]
        formats = {}
        for db_format in self.formats:
            for subfolder in ['', 'databases/']:
                db_format_dir = path + '/%s' % subfolder + db_format
                if not self.config.dev and not isdir(db_format_dir):
                    continue
                formats[db_format] = db_format_dir
                break
        return formats

    def set_wol(self) -> None:
        wol_dir = self.config.databases[self.database]
        self.paths[self.database] = wol_dir
        self.builds[self.database] = self.get_db_formats()
        # WOL is a fasta file used to build indices
        fna_dir = '%s/genomes' % wol_dir
        genomes = glob.glob('%s/fna/*' % fna_dir)
        if not genomes:
            metadata_fp = '%s/wol/metadata.tsv' % wol_dir
            batch_down_fp = '%s/wol/batch_down.sh' % RESOURCES
            make_down_list_py = '%s/wol/make_down_list.py' % RESOURCES
            cmd = 'mkdir -p %s\n' % fna_dir
            cmd += 'cd %s\n' % fna_dir
            cmd += '%s %s > download.list\n' % (make_down_list_py, metadata_fp)
            cmd += 'bash %s download.list' % batch_down_fp
            self.cmds['download_wol_genomes'] = [cmd]
        else:
            self.wol['fna'] = genomes

        lineages_fp = '%s/wol/lineages.txt' % RESOURCES
        lineages_pd = pd.read_csv(lineages_fp, header=None, sep='\t')
        lineages_pd.columns = ['Genome ID', 'taxonomy']
        lineages_pd['species'] = [x.split(';')[-1].strip().replace(
            ' ', '_').replace( '[', '').replace(
            ']', '') for x in lineages_pd['taxonomy']]
        self.wol['taxonomy'] = lineages_pd

        tree_fp = '%s/wol/tree.nwk' % RESOURCES
        self.wol['tree'] = TreeNode.read(tree_fp, format='newick')

        sizes_fp = '%s/wol/genome_sizes.txt' % RESOURCES
        sizes = pd.read_csv(sizes_fp, header=None, sep='\t')
        sizes.columns = ['gid', 'length']
        self.wol['sizes'] = sizes
        self.register_command()

    def set_pfam(self) -> None:
        pfams_dir = self.config.databases[self.database]
        self.paths[self.database] = pfams_dir
        self.pfams['dir'] = pfams_dir
        cmd = get_pfam_wget_cmd(pfams_dir)
        if cmd:
            self.cmds[''] = [cmd]
        else:
            self.hmms_pd = get_hmm_dat(pfams_dir)
        self.register_command()
        self.pfam_models()

    def pfam_models(self):
        pfam_terms_dir = '%s/pfam_terms' % RESOURCES
        self.pfams['res'] = pfam_terms_dir
        mkdr(pfam_terms_dir)
        if self.config.show_pfams or self.config.purge_pfams:
            term_dirs = sorted(glob.glob('%s/*' % pfam_terms_dir))
            if self.config.show_pfams:
                if term_dirs:
                    print(' > already extracted Pfam models', end='')
                    if self.config.purge_pfams:
                        print(' and removed:')
                    else:
                        print(':')
                    for term_dir in term_dirs:
                        hmm_fps = glob.glob('%s/*.hmm' % term_dir)
                        print('   - %s (%s HMMs)' % (term_dir, len(hmm_fps)))
                        if self.config.purge_pfams:
                            for hmm_fp in hmm_fps:
                                os.remove(hmm_fp)
                            os.rmdir(term_dir)
                else:
                    print(' > No Pfam models previously extracted')

    def get_pfam_terms(self, params):
        self.database = 'pfam'
        terms = {}
        hmm = '%s/Pfam-A.hmm' % self.pfams['dir']
        for term in params['terms']:
            term_pd = self.hmms_pd.loc[
                self.hmms_pd['DE'].str.lower().str.contains(term.lower())
            ].copy()
            if not term_pd.shape[0]:
                print('[hmmer] No Pfam model for term "%s" (ignored)' % term)
                continue
            pfams, cmd = get_pfams_cmd(hmm, term_pd, term, self.pfams['res'])
            terms[term] = pfams
            self.cmds[term] = [cmd]
        # collect the "pfams per term" for the parameters of the current tools
        params['terms'] = terms
        # add these "pfams per term" to a full, database-level collection
        self.pfams['terms'].update(terms)
        self.register_command()

    def get_dbcan_hmms(self) -> None:
        """Get all the .hmm files from the dbCAN database."""
        for root, _, files in os.walk(self.config.databases['dbcan']):
            for fil in files:
                if fil.endswith('.hmm'):
                    hmm = root + '/' + fil
                    rad = splitext(hmm)[0]
                    fas, dia = '%s.fas' % rad, '%s.dmnd' % rad
                    cmd = 'diamond makedb --in %s -d %s\n' % (fas, dia)
                    if 'cazy' not in self.hmms_dias:
                        self.hmms_dias['cazy'] = {}
                    self.hmms_dias['cazy'][basename(rad)] = [hmm, dia]
                    self.cmds.setdefault(basename(rad), []).append(cmd)

    def set_dbcan(self) -> None:
        self.get_dbcan_hmms()
        m = '%s/dbCAN-seq/metadata.txt' % self.config.databases['dbcan']
        if isfile(m):
            self.dbcan_meta = pd.read_csv(m, header=0, index_col=0, sep='\t')

    def set_shogun(self) -> None:
        shogun_config = self.config.databases['shogun']
        self.paths[self.database] = shogun_config
        for k, v in shogun_config.items():
            if k == 'rep82':
                if not isdir(v) or not glob.glob('%s/*' % v):
                    cmd = 'mkdir -p %s\ncd %s\n' % (v, v)
                    cmd += 'wget -i https://raw.githubusercontent.com/knights' \
                           '-lab/SHOGUN/master/docs/shogun_db_links.txt\n'
                    self.cmds['download_shogun'] = [cmd]


